# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

from flask import make_response

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import tornado.web
import flask
import flask.json
import flask_login
import octoprint.vendor.flask_principal as flask_principal
import flask_assets
import webassets.updater
import webassets.utils
import functools
import time
import uuid
import threading
import logging
import netaddr
import os
import io
import collections

from octoprint.settings import settings
from octoprint.util import deprecated
import octoprint.server
import octoprint.access.users
import octoprint.plugin

from octoprint.util import DefaultOrderedDict
from octoprint.util.json import JsonEncoding
from octoprint.util.net import is_lan_address

from octoprint.events import eventManager, Events

from werkzeug.local import LocalProxy
from cachelib import BaseCache

from past.builtins import basestring, long

try:
	from os import scandir, walk
except ImportError:
	from scandir import scandir, walk

#~~ monkey patching

def enable_additional_translations(default_locale="en", additional_folders=None):
	import os
	from flask import _request_ctx_stack
	from babel import support, Locale
	import flask_babel

	if additional_folders is None:
		additional_folders = []

	logger = logging.getLogger(__name__)

	def fixed_list_translations(self):
		"""Returns a list of all the locales translations exist for.  The
		list returned will be filled with actual locale objects and not just
		strings.
		"""
		def list_translations(dirname):
			if not os.path.isdir(dirname):
				return []
			result = []
			for entry in scandir(dirname):
				locale_dir = os.path.join(entry.path, 'LC_MESSAGES')
				if not os.path.isdir(locale_dir):
					continue
				if any(filter(lambda x: x.name.endswith('.mo'), scandir(locale_dir))):
					result.append(Locale.parse(entry.name))
			return result

		dirs = additional_folders + [os.path.join(self.app.root_path, 'translations')]

		# translations from plugins
		plugins = octoprint.plugin.plugin_manager().enabled_plugins
		for name, plugin in plugins.items():
			plugin_translation_dir = os.path.join(plugin.location, 'translations')
			if not os.path.isdir(plugin_translation_dir):
				continue
			dirs.append(plugin_translation_dir)

		result = [Locale.parse(default_locale)]

		for dir in dirs:
			result += list_translations(dir)
		return result

	def fixed_get_translations():
		"""Returns the correct gettext translations that should be used for
		this request.  This will never fail and return a dummy translation
		object if used outside of the request or if a translation cannot be
		found.
		"""
		ctx = _request_ctx_stack.top
		if ctx is None:
			return None
		translations = getattr(ctx, 'babel_translations', None)
		if translations is None:
			locale = flask_babel.get_locale()
			translations = support.Translations()

			if str(locale) != default_locale:
				# plugin translations
				plugins = octoprint.plugin.plugin_manager().enabled_plugins
				for name, plugin in plugins.items():
					dirs = list(map(lambda x: os.path.join(x, "_plugins", name), additional_folders)) + [os.path.join(plugin.location, 'translations')]
					for dirname in dirs:
						if not os.path.isdir(dirname):
							continue

						try:
							plugin_translations = support.Translations.load(dirname, [locale])
						except Exception:
							logger.exception("Error while trying to load translations for plugin {name}".format(**locals()))
						else:
							if isinstance(plugin_translations, support.Translations):
								translations = translations.merge(plugin_translations)
								logger.debug("Using translation plugin folder {dirname} from plugin {name} for locale {locale}".format(**locals()))
								break
					else:
						logger.debug("No translations for locale {locale} from plugin {name}".format(**locals()))

				# core translations
				dirs = additional_folders + [os.path.join(ctx.app.root_path, 'translations')]
				for dirname in dirs:
					core_translations = support.Translations.load(dirname, [locale])
					if isinstance(core_translations, support.Translations):
						logger.debug("Using translation core folder {dirname} for locale {locale}".format(**locals()))
						break
				else:
					logger.debug("No translations for locale {} in core folders".format(locale))
				translations = translations.merge(core_translations)

			ctx.babel_translations = translations
		return translations

	flask_babel.Babel.list_translations = fixed_list_translations
	flask_babel.get_translations = fixed_get_translations

def fix_webassets_cache():
	from webassets import cache

	error_logger = logging.getLogger(__name__ + ".fix_webassets_cache")

	def fixed_set(self, key, data):
		import os
		import tempfile
		import pickle
		import shutil

		if not os.path.exists(self.directory):
			error_logger.warning("Cache directory {} doesn't exist, not going "
			                     "to attempt to write cache file".format(self.directory))

		md5 = '%s' % cache.make_md5(self.V, key)
		filename = os.path.join(self.directory, md5)
		fd, temp_filename = tempfile.mkstemp(prefix='.' + md5,
		                                     dir=self.directory)
		try:
			with os.fdopen(fd, 'wb') as f:
				pickle.dump(data, f)
				f.flush()
			shutil.move(temp_filename, filename)
		except Exception:
			os.remove(temp_filename)
			raise

	def fixed_get(self, key):
		import os
		import errno
		import warnings
		from webassets.cache import make_md5

		if not os.path.exists(self.directory):
			error_logger.warning("Cache directory {} doesn't exist, not going "
			                     "to attempt to read cache file".format(self.directory))
			return None

		try:
			hash = make_md5(self.V, key)
		except IOError as e:
			if e.errno != errno.ENOENT:
				raise
			return None

		filename = os.path.join(self.directory, '%s' % hash)
		try:
			f = io.open(filename, 'rb')
		except IOError as e:
			if e.errno != errno.ENOENT:
				error_logger.exception("Got an exception while trying to open webasset file {}".format(filename))
			return None
		try:
			result = f.read()
		finally:
			f.close()

		unpickled = webassets.cache.safe_unpickle(result)
		if unpickled is None:
			warnings.warning('Ignoring corrupted cache file %s' % filename)
		return unpickled

	cache.FilesystemCache.set = fixed_set
	cache.FilesystemCache.get = fixed_get

def fix_webassets_filtertool():
	from webassets.merge import FilterTool, log, MemoryHunk

	error_logger = logging.getLogger(__name__ + ".fix_webassets_filtertool")

	def fixed_wrap_cache(self, key, func):
		"""Return cache value ``key``, or run ``func``.
		"""
		if self.cache:
			if not self.no_cache_read:
				log.debug('Checking cache for key %s', key)
				content = self.cache.get(key)
				if not content in (False, None):
					log.debug('Using cached result for %s', key)
					return MemoryHunk(content)

		try:
			content = func().getvalue()
			if self.cache:
				try:
					log.debug('Storing result in cache with key %s', key,)
					self.cache.set(key, content)
				except Exception:
					error_logger.exception("Got an exception while trying to save file to cache, not caching")
			return MemoryHunk(content)
		except Exception:
			error_logger.exception("Got an exception while trying to apply filter, ignoring file")
			return MemoryHunk("")

	FilterTool._wrap_cache = fixed_wrap_cache

def fix_flask_jsonify():
	def fixed_jsonify(*args, **kwargs):
		"""Backported from https://github.com/pallets/flask/blob/7e714bd28b6e96d82b2848b48cf8ff48b517b09b/flask/json/__init__.py#L257"""
		from flask.json import current_app, dumps

		indent = None
		separators = (',', ':')

		if current_app.config['JSONIFY_PRETTYPRINT_REGULAR'] or current_app.debug:
			indent = 2
			separators = (', ', ': ')

		if args and kwargs:
			raise TypeError('jsonify() behavior undefined when passed both args and kwargs')
		elif len(args) == 1:  # single args are passed directly to dumps()
			data = args[0]
		else:
			data = args or kwargs

		return current_app.response_class(
			dumps(data, indent=indent, separators=separators, allow_nan=False) + '\n',
			mimetype='application/json'
		)

	flask.jsonify = fixed_jsonify

#~~ WSGI environment wrapper for reverse proxying

class ReverseProxiedEnvironment(object):

	@staticmethod
	def to_header_candidates(values):
		if values is None:
			return []
		if not isinstance(values, (list, tuple)):
			values = [values]
		to_wsgi_format = lambda header: "HTTP_" + header.upper().replace("-", "_")
		return list(map(to_wsgi_format, values))

	@staticmethod
	def valid_ip(address):
		import netaddr
		try:
			netaddr.IPAddress(address)
			return True
		except Exception:
			return False

	def __init__(self,
	             header_prefix=None,
	             header_scheme=None,
	             header_host=None,
	             header_server=None,
	             header_port=None,
	             prefix=None,
	             scheme=None,
	             host=None,
	             server=None,
	             port=None):

		# sensible defaults
		if header_prefix is None:
			header_prefix = ["x-script-name"]
		if header_scheme is None:
			header_scheme = ["x-forwarded-proto", "x-scheme"]
		if header_host is None:
			header_host = ["x-forwarded-host"]
		if header_server is None:
			header_server = ["x-forwarded-server"]
		if header_port is None:
			header_port = ["x-forwarded-port"]

		# header candidates
		self._headers_prefix = self.to_header_candidates(header_prefix)
		self._headers_scheme = self.to_header_candidates(header_scheme)
		self._headers_host = self.to_header_candidates(header_host)
		self._headers_server = self.to_header_candidates(header_server)
		self._headers_port = self.to_header_candidates(header_port)

		# fallback prefix & scheme & host from config
		self._fallback_prefix = prefix
		self._fallback_scheme = scheme
		self._fallback_host = host
		self._fallback_server = server
		self._fallback_port = port

	def __call__(self, environ):
		def retrieve_header(header_type):
			candidates = getattr(self, "_headers_" + header_type, [])
			fallback = getattr(self, "_fallback_" + header_type, None)

			for candidate in candidates:
				value = environ.get(candidate, None)
				if value is not None:
					return value
			else:
				return fallback

		def host_to_server_and_port(host, scheme):
			if host is None:
				return None, None

			default_port = "443" if scheme == "https" else "80"
			host = host.strip()

			if ":" in host:
				# we might have an ipv6 address here, or a port, or both

				if host[0] == "[":
					# that looks like an ipv6 address with port, e.g. [fec1::1]:80
					address_end = host.find("]")
					if address_end == -1:
						# no ], that looks like a seriously broken address
						return None, None

					# extract server ip, skip enclosing [ and ]
					server = host[1:address_end]
					tail = host[address_end + 1:]

					# now check if there's also a port
					if len(tail) and tail[0] == ":":
						# port included as well
						port = tail[1:]
					else:
						# no port, use default one
						port = default_port

				elif self.__class__.valid_ip(host):
					# ipv6 address without port
					server = host
					port = default_port

				else:
					# ipv4 address with port
					server, port = host.rsplit(":", 1)

			else:
				server = host
				port = default_port

			return server, port

		# determine prefix
		prefix = retrieve_header("prefix")
		if prefix is not None:
			environ["SCRIPT_NAME"] = prefix
			path_info = environ["PATH_INFO"]
			if path_info.startswith(prefix):
				environ["PATH_INFO"] = path_info[len(prefix):]

		# determine scheme
		scheme = retrieve_header("scheme")
		if scheme is not None and "," in scheme:
			# Scheme might be something like "https,https" if doubly-reverse-proxied
			# without stripping original scheme header first, make sure to only use
			# the first entry in such a case. See #1391.
			scheme, _ = map(lambda x: x.strip(), scheme.split(",", 1))
		if scheme is not None:
			environ["wsgi.url_scheme"] = scheme

		# determine host
		url_scheme = environ["wsgi.url_scheme"]
		host = retrieve_header("host")
		if host is not None:
			# if we have a host, we take server_name and server_port from it
			server, port = host_to_server_and_port(host, url_scheme)
			environ["HTTP_HOST"] = host
			environ["SERVER_NAME"] = server
			environ["SERVER_PORT"] = port

		elif environ.get("HTTP_HOST", None) is not None:
			# if we have a Host header, we use that and make sure our server name and port properties match it
			host = environ["HTTP_HOST"]
			server, port = host_to_server_and_port(host, url_scheme)
			environ["SERVER_NAME"] = server
			environ["SERVER_PORT"] = port

		else:
			# else we take a look at the server and port headers and if we have
			# something there we derive the host from it

			# determine server - should usually not be used
			server = retrieve_header("server")
			if server is not None:
				environ["SERVER_NAME"] = server

			# determine port - should usually not be used
			port = retrieve_header("port")
			if port is not None:
				environ["SERVER_PORT"] = port

			# reconstruct host header
			if url_scheme == "http" and environ["SERVER_PORT"] == "80" or url_scheme == "https" and environ["SERVER_PORT"] == "443":
				# default port for scheme, can be skipped
				environ["HTTP_HOST"] = environ["SERVER_NAME"]
			else:
				server_name_component = environ["SERVER_NAME"]
				if ":" in server_name_component and self.__class__.valid_ip(server_name_component):
					# this is an ipv6 address, we need to wrap that in [ and ] before appending the port
					server_name_component = "[" + server_name_component + "]"

				environ["HTTP_HOST"] = server_name_component + ":" + environ["SERVER_PORT"]

		# call wrapped app with rewritten environment
		return environ

#~~ request and response versions

from werkzeug.utils import cached_property

class OctoPrintFlaskRequest(flask.Request):
	environment_wrapper = staticmethod(lambda x: x)

	def __init__(self, environ, *args, **kwargs):
		# apply environment wrapper to provided WSGI environment
		flask.Request.__init__(self, self.environment_wrapper(environ), *args, **kwargs)

	@cached_property
	def cookies(self):
		# strip cookie_suffix from all cookies in the request, return result
		cookies = flask.Request.cookies.__get__(self)

		result = dict()
		desuffixed = dict()
		for key, value in cookies.items():
			if key.endswith(self.cookie_suffix):
				desuffixed[key[:-len(self.cookie_suffix)]] = value
			else:
				result[key] = value

		result.update(desuffixed)
		return result

	@cached_property
	def server_name(self):
		"""Short cut to the request's server name header"""
		return self.environ.get("SERVER_NAME")

	@cached_property
	def server_port(self):
		"""Short cut to the request's server port header"""
		return self.environ.get("SERVER_PORT")

	@cached_property
	def cookie_suffix(self):
		"""
		Request specific suffix for set and read cookies

		We need this because cookies are not port-specific and we don't want to overwrite our
		session and other cookies from one OctoPrint instance on our machine with those of another
		one who happens to listen on the same address albeit a different port or script root.
		"""
		result = "_P" + self.server_port
		if self.script_root:
			return result + "_R" + self.script_root.replace("/", "|")
		return result


class OctoPrintFlaskResponse(flask.Response):
	def set_cookie(self, key, *args, **kwargs):
		# restrict cookie path to script root
		kwargs["path"] = flask.request.script_root + kwargs.get("path", "/")

		# add request specific cookie suffix to name
		flask.Response.set_cookie(self, key + flask.request.cookie_suffix, *args, **kwargs)

	def delete_cookie(self, key, path='/', domain=None):
		flask.Response.delete_cookie(self, key, path=path, domain=domain)

		# we also still might have a cookie left over from before we started prefixing, delete that manually
		# without any pre processing (no path prefix, no key suffix)
		flask.Response.set_cookie(self, key, expires=0, max_age=0, path=path, domain=domain)


class OctoPrintSessionInterface(flask.sessions.SecureCookieSessionInterface):

	def save_session(self, app, session, response):
		if flask.g.get("login_via_apikey", False):
			return
		return super(OctoPrintSessionInterface, self).save_session(app, session, response)


#~~ passive login helper

_cached_local_networks = None
def _local_networks():
	global _cached_local_networks

	if _cached_local_networks is None:
		logger = logging.getLogger(__name__)
		local_networks = netaddr.IPSet([])
		for entry in settings().get(["accessControl", "localNetworks"]):
			network = netaddr.IPNetwork(entry)
			local_networks.add(network)
			logger.debug("Added network {} to localNetworks".format(network))

			if network.version == 4:
				network_v6 = network.ipv6()
				local_networks.add(network_v6)
				logger.debug("Also added v6 representation of v4 network {} = {} to localNetworks".format(network, network_v6))

		_cached_local_networks = local_networks

	return _cached_local_networks


def passive_login():
	logger = logging.getLogger(__name__)

	user = flask_login.current_user

	remote_address = get_remote_address(flask.request)
	ip_check_enabled = settings().getBoolean(["server", "ipCheck", "enabled"])
	ip_check_trusted = settings().get(["server", "ipCheck", "trustedSubnets"])

	if isinstance(user, LocalProxy):
		# noinspection PyProtectedMember
		user = user._get_current_object()

	def login(u):
		# login known user
		if not u.is_anonymous:
			u = octoprint.server.userManager.login_user(u)
		flask_login.login_user(u)
		flask_principal.identity_changed.send(flask.current_app._get_current_object(),
		                                      identity=flask_principal.Identity(u.get_id()))
		if hasattr(u, "session"):
			flask.session["usersession.id"] = u.session
		flask.g.user = u

		eventManager().fire(Events.USER_LOGGED_IN, payload=dict(username=u.get_id()))

		return u

	def determine_user(u):
		if not u.is_anonymous and u.is_active:
			# known active user
			logger.info("Passively logging in user {} from {}".format(u.get_id(), remote_address))

		elif settings().getBoolean(["accessControl", "autologinLocal"]) \
				and settings().get(["accessControl", "autologinAs"]) is not None \
				and settings().get(["accessControl", "localNetworks"]) is not None \
				and not "active_logout" in flask.request.cookies \
				and remote_address:
			# attempt local autologin
			autologin_as = settings().get(["accessControl", "autologinAs"])
			local_networks = _local_networks()
			logger.debug("Checking if remote address {} is in localNetworks ({!r})".format(remote_address, local_networks))

			try:
				if netaddr.IPAddress(remote_address) in local_networks:
					autologin_user = octoprint.server.userManager.find_user(autologin_as)
					if autologin_user is not None and autologin_user.is_active:
						logger.info("Passively logging in user {} from {} via autologin".format(autologin_as, remote_address))
						return autologin_user
			except Exception:
				logger.exception("Could not autologin user {} from {} for networks {}".format(autologin_as, remote_address, local_networks))

		if not u.is_active:
			# inactive user, switch to anonymous
			u = octoprint.server.userManager.anonymous_user_factory()

		return u

	user = login(determine_user(user))
	response = user.as_dict()
	response["_is_external_client"] = ip_check_enabled and not is_lan_address(remote_address,
	                                                                          additional_private=ip_check_trusted)
	return flask.jsonify(response)


#~~ cache decorator for cacheable views

class LessSimpleCache(BaseCache):
	"""
	Slightly improved version of :class:`SimpleCache`.

	Setting ``default_timeout`` or ``timeout`` to ``-1`` will have no timeout be applied at all.
	"""

	def __init__(self, threshold=500, default_timeout=300):
		BaseCache.__init__(self, default_timeout=default_timeout)
		self._mutex = threading.RLock()
		self._cache = {}
		self._bypassed = set()
		self.clear = self._cache.clear
		self._threshold = threshold

	def _prune(self):
		if self.over_threshold():
			now = time.time()
			for idx, (key, (expires, _)) in enumerate(self._cache.items()):
				if expires is not None and expires <= now or idx % 3 == 0:
					with self._mutex:
						self._cache.pop(key, None)

	def get(self, key):
		import pickle
		now = time.time()
		with self._mutex:
			expires, value = self._cache.get(key, (0, None))
		if expires is None or expires > now:
			return pickle.loads(value)

	def set(self, key, value, timeout=None):
		import pickle

		with self._mutex:
			self._prune()
			self._cache[key] = (self.calculate_timeout(timeout=timeout),
								pickle.dumps(value, pickle.HIGHEST_PROTOCOL))
			if key in self._bypassed:
				self._bypassed.remove(key)

	def add(self, key, value, timeout=None):
		with self._mutex:
			self.set(key, value, timeout=None)
			self._cache.setdefault(key, self._cache[key])

	def delete(self, key):
		with self._mutex:
			self._cache.pop(key, None)

	def calculate_timeout(self, timeout=None):
		if timeout is None:
			timeout = self.default_timeout
		if timeout == -1:
			return None
		return time.time() + timeout

	def over_threshold(self):
		if self._threshold is None:
			return False
		with self._mutex:
			return len(self._cache) > self._threshold

	def __getitem__(self, key):
		return self.get(key)

	def __setitem__(self, key, value):
		return self.set(key, value)

	def __delitem__(self, key):
		return self.delete(key)

	def __contains__(self, key):
		with self._mutex:
			return key in self._cache

	def set_bypassed(self, key):
		with self._mutex:
			self._bypassed.add(key)

	def is_bypassed(self, key):
		with self._mutex:
			return key in self._bypassed

_cache = LessSimpleCache()

def cached(timeout=5 * 60, key=lambda: "view:%s" % flask.request.path, unless=None, refreshif=None, unless_response=None):
	def decorator(f):
		@functools.wraps(f)
		def decorated_function(*args, **kwargs):
			logger = logging.getLogger(__name__)

			cache_key = key()

			def f_with_duration(*args, **kwargs):
				start_time = time.time()
				try:
					return f(*args, **kwargs)
				finally:
					elapsed = time.time() - start_time
					logger.debug(
						"Needed {elapsed:.2f}s to render {path} (key: {key})".format(elapsed=elapsed,
						                                                             path=flask.request.path,
						                                                             key=cache_key))

			# bypass the cache if "unless" condition is true
			if callable(unless) and unless():
				logger.debug("Cache for {path} bypassed, calling wrapped function".format(path=flask.request.path))
				_cache.set_bypassed(cache_key)
				return f_with_duration(*args, **kwargs)

			# also bypass the cache if it's disabled completely
			if not settings().getBoolean(["devel", "cache", "enabled"]):
				logger.debug("Cache for {path} disabled, calling wrapped function".format(path=flask.request.path))
				_cache.set_bypassed(cache_key)
				return f_with_duration(*args, **kwargs)

			rv = _cache.get(cache_key)

			# only take the value from the cache if we are not required to refresh it from the wrapped function
			if rv is not None and (not callable(refreshif) or not refreshif(rv)):
				logger.debug("Serving entry for {path} from cache (key: {key})".format(path=flask.request.path, key=cache_key))
				if not "X-From-Cache" in rv.headers:
					rv.headers["X-From-Cache"] = "true"
				return rv

			# get value from wrapped function
			logger.debug("No cache entry or refreshing cache for {path} (key: {key}), calling wrapped function".format(path=flask.request.path, key=cache_key))
			rv = f_with_duration(*args, **kwargs)

			# do not store if the "unless_response" condition is true
			if callable(unless_response) and unless_response(rv):
				logger.debug("Not caching result for {path} (key: {key}), bypassed".format(path=flask.request.path, key=cache_key))
				_cache.set_bypassed(cache_key)
				return rv

			# store it in the cache
			_cache.set(cache_key, rv, timeout=timeout)

			return rv

		return decorated_function

	return decorator

def is_in_cache(key=lambda: "view:%s" % flask.request.path):
	if callable(key):
		key = key()
	return key in _cache

def is_cache_bypassed(key=lambda: "view:%s" % flask.request.path):
	if callable(key):
		key = key()
	return _cache.is_bypassed(key)

def cache_check_headers():
	return "no-cache" in flask.request.cache_control or "no-cache" in flask.request.pragma

def cache_check_response_headers(response):
	if not isinstance(response, flask.Response):
		return False

	headers = response.headers

	if "Cache-Control" in headers and ("no-cache" in headers["Cache-Control"] or "no-store" in headers["Cache-Control"]):
		return True

	if "Pragma" in headers and "no-cache" in headers["Pragma"]:
		return True

	if "Expires" in headers and headers["Expires"] in ("0", "-1"):
		return True

	return False

def cache_check_status_code(response, valid):
	if not isinstance(response, flask.Response):
		return False

	if callable(valid):
		return not valid(response.status_code)
	else:
		return response.status_code not in valid

class PreemptiveCache(object):

	def __init__(self, cachefile):
		self.cachefile = cachefile
		self.environment = None

		self._logger = logging.getLogger(__name__ + "." + self.__class__.__name__)

		self._lock = threading.RLock()

	def record(self, data, unless=None, root=None):
		if callable(unless) and unless():
			return

		entry_data = data
		if callable(entry_data):
			entry_data = entry_data()

		if entry_data is not None:
			if root is None:
				from flask import request
				root = request.path
			self.add_data(root, entry_data)

	def has_record(self, data, root=None):
		if callable(data):
			data = data()

		if data is None:
			return False

		if root is None:
			from flask import request
			root = request.path

		all_data = self.get_data(root)
		for existing in all_data:
			if self._compare_data(data, existing):
				return True

		return False

	def clean_all_data(self, cleanup_function):
		assert callable(cleanup_function)

		with self._lock:
			all_data = self.get_all_data()
			for root, entries in list(all_data.items()):
				old_count = len(entries)
				entries = cleanup_function(root, entries)
				if not entries:
					del all_data[root]
					self._logger.debug("Removed root {} from preemptive cache".format(root))
				elif len(entries) < old_count:
					all_data[root] = entries
					self._logger.debug("Removed {} entries from preemptive cache for root {}".format(old_count - len(entries), root))
			self.set_all_data(all_data)

		return all_data

	def get_all_data(self):
		import yaml

		cache_data = None
		with self._lock:
			try:
				with io.open(self.cachefile, 'rt') as f:
					cache_data = yaml.safe_load(f)
			except IOError as e:
				import errno
				if e.errno != errno.ENOENT:
					raise
			except Exception:
				self._logger.exception("Error while reading {}".format(self.cachefile))

		if cache_data is None:
			cache_data = dict()

		if not self._validate_data(cache_data):
			self._logger.warn("Preemptive cache data was invalid, ignoring it")
			cache_data = dict()

		return cache_data

	def get_data(self, root):
		cache_data = self.get_all_data()
		return cache_data.get(root, list())

	def set_all_data(self, data):
		from octoprint.util import atomic_write
		import yaml

		with self._lock:
			try:
				with atomic_write(self.cachefile, 'wt', max_permissions=0o666) as handle:
					yaml.safe_dump(data, handle,default_flow_style=False, indent=4, allow_unicode=True)
			except Exception:
				self._logger.exception("Error while writing {}".format(self.cachefile))

	def set_data(self, root, data):
		with self._lock:
			all_data = self.get_all_data()
			all_data[root] = data
			self.set_all_data(all_data)

	def add_data(self, root, data):
		def split_matched_and_unmatched(entry, entries):
			matched = []
			unmatched = []

			for e in entries:
				if self._compare_data(e, entry):
					matched.append(e)
				else:
					unmatched.append(e)

			return matched, unmatched

		with self._lock:
			cache_data = self.get_all_data()

			if not root in cache_data:
				cache_data[root] = []

			existing, other = split_matched_and_unmatched(data, cache_data[root])

			def get_newest(entries):
				result = None
				for entry in entries:
					if "_timestamp" in entry and (result is None or ("_timestamp" in result and result["_timestamp"] < entry["_timestamp"])):
						result = entry
				return result

			to_persist = get_newest(existing)
			if not to_persist:
				import copy
				to_persist = copy.deepcopy(data)
				to_persist["_timestamp"] = time.time()
				to_persist["_count"] = 1
				self._logger.info("Adding entry for {} and {!r}".format(root, to_persist))
			else:
				to_persist["_timestamp"] = time.time()
				to_persist["_count"] = to_persist.get("_count", 0) + 1
				self._logger.debug("Updating timestamp and counter for {} and {!r}".format(root, data))

			self.set_data(root, [to_persist] + other)

	def _compare_data(self, a, b):
		from octoprint.util import dict_filter

		def strip_ignored(d):
			return dict_filter(d, lambda k, v: not k.startswith("_"))

		return set(strip_ignored(a).items()) == set(strip_ignored(b).items())

	def _validate_data(self, data):
		if not isinstance(data, dict):
			return False

		for root, entries in data.items():
			if not isinstance(entries, list):
				return False

			for entry in entries:
				if not self._validate_entry(entry):
					return False

		return True

	def _validate_entry(self, entry):
		return isinstance(entry, dict) and "_timestamp" in entry and "_count" in entry


def preemptively_cached(cache, data, unless=None):
	def decorator(f):
		@functools.wraps(f)
		def decorated_function(*args, **kwargs):
			try:
				cache.record(data, unless=unless)
			except Exception:
				logging.getLogger(__name__).exception("Error while recording preemptive cache entry: {!r}".format(data))
			return f(*args, **kwargs)
		return decorated_function
	return decorator


def etagged(etag):
	def decorator(f):
		@functools.wraps(f)
		def decorated_function(*args, **kwargs):
			rv = f(*args, **kwargs)
			if isinstance(rv, flask.Response):
				try:
					result = etag
					if callable(result):
						result = result(rv)
					if result:
						rv.set_etag(result)
				except Exception:
					logging.getLogger(__name__).exception("Error while calculating the etag value for response {!r}".format(rv))
			return rv
		return decorated_function
	return decorator


def lastmodified(date):
	def decorator(f):
		@functools.wraps(f)
		def decorated_function(*args, **kwargs):
			rv = f(*args, **kwargs)
			if not "Last-Modified" in rv.headers:
				try:
					result = date
					if callable(result):
						result = result(rv)

					if not isinstance(result, basestring):
						from werkzeug.http import http_date
						result = http_date(result)

					if result:
						rv.headers["Last-Modified"] = result
				except Exception:
					logging.getLogger(__name__).exception("Error while calculating the lastmodified value for response {!r}".format(rv))
			return rv
		return decorated_function
	return decorator


def conditional(condition, met):
	def decorator(f):
		@functools.wraps(f)
		def decorated_function(*args, **kwargs):
			try:
				if callable(condition) and condition():
					# condition has been met, return met-response
					rv = met
					if callable(met):
						rv = met()
					return rv
			except Exception:
				logging.getLogger(__name__).exception("Error while evaluating conditional {!r} or met {!r}".format(condition, met))

			# condition hasn't been met, call decorated function
			return f(*args, **kwargs)
		return decorated_function
	return decorator


def with_client_revalidation(f):
	@functools.wraps(f)
	def decorated_function(*args, **kwargs):
		r = f(*args, **kwargs)

		if isinstance(r, flask.Response):
			r = add_revalidation_response_headers(r)

		return r
	return decorated_function


def with_revalidation_checking(etag_factory=None,
                               lastmodified_factory=None,
                               condition=None,
                               unless=None):
	if etag_factory is None:
		def etag_factory(lm=None):
			return None

	if lastmodified_factory is None:
		def lastmodified_factory():
			return None

	if condition is None:
		def condition(lm=None, etag=None):
			if lm is None:
				lm = lastmodified_factory()

			if etag is None:
				etag = etag_factory(lm=lm)

			return check_lastmodified(lm) and check_etag(etag)

	if unless is None:
		def unless():
			return False

	def decorator(f):
		@functools.wraps(f)
		def decorated_function(*args, **kwargs):
			lm = lastmodified_factory()
			etag = etag_factory(lm)

			if condition(lm, etag) and not unless():
				return make_response("Not Modified", 304)

			# generate response
			response = f(*args, **kwargs)

			# set etag header if not already set
			if etag and response.get_etag()[0] is None:
				response.set_etag(etag)

			# set last modified header if not already set
			if lm and response.headers.get("Last-Modified", None) is None:
				if not isinstance(lm, basestring):
					from werkzeug.http import http_date
					lm = http_date(lm)
				response.headers["Last-Modified"] = lm

			response = add_no_max_age_response_headers(response)
			return response
		return decorated_function
	return decorator


def check_etag(etag):
	if etag is None:
		return False

	return flask.request.method in ("GET", "HEAD") and \
	       flask.request.if_none_match is not None and \
	       etag in flask.request.if_none_match


def check_lastmodified(lastmodified):
	if lastmodified is None:
		return False

	from datetime import datetime
	if isinstance(lastmodified, (int, long, float)):
		# max(86400, lastmodified) is workaround for https://bugs.python.org/issue29097,
		# present in CPython 3.6.x up to 3.7.1.
		#
		# I think it's fair to say that we'll never encounter lastmodified values older than
		# 1970-01-02 so this is a safe workaround.
		lastmodified = datetime.fromtimestamp(max(86400, lastmodified)).replace(microsecond=0)

	if not isinstance(lastmodified, datetime):
		raise ValueError("lastmodified must be a datetime or float or int instance but, got {} instead".format(lastmodified.__class__))

	return flask.request.method in ("GET", "HEAD") and \
	       flask.request.if_modified_since is not None and \
	       lastmodified >= flask.request.if_modified_since


def add_revalidation_response_headers(response):
	import werkzeug.http

	cache_control = werkzeug.http.parse_dict_header(response.headers.get("Cache-Control", ""))
	if "no-cache" not in cache_control:
		cache_control["no-cache"] = None
	if "must-revalidate" not in cache_control:
		cache_control["must-revalidate"] = None
	response.headers["Cache-Control"] = werkzeug.http.dump_header(cache_control)

	return response


def add_non_caching_response_headers(response):
	import werkzeug.http

	cache_control = werkzeug.http.parse_dict_header(response.headers.get("Cache-Control", ""))
	if "no-store" not in cache_control:
		cache_control["no-store"] = None
	if "no-cache" not in cache_control:
		cache_control["no-cache"] = None
	if "must-revalidate" not in cache_control:
		cache_control["must-revalidate"] = None
	if "post-check" not in cache_control or cache_control["post-check"] != "0":
		cache_control["post-check"] = "0"
	if "pre-check" not in cache_control or cache_control["pre-check"] != "0":
		cache_control["pre-check"] = "0"
	if "max-age" not in cache_control or cache_control["max-age"] != "0":
		cache_control["max-age"] = "0"
	response.headers["Cache-Control"] = werkzeug.http.dump_header(cache_control)

	response.headers["Pragma"] = "no-cache"
	response.headers["Expires"] = "-1"
	return response


def add_no_max_age_response_headers(response):
	import werkzeug.http

	cache_control = werkzeug.http.parse_dict_header(response.headers.get("Cache-Control", ""))
	if "max-age" not in cache_control or cache_control["max-age"] != "0":
		cache_control["max-age"] = "0"
	response.headers["Cache-Control"] = werkzeug.http.dump_header(cache_control)

	return response


#~~ access validators for use with tornado

def permission_validator(request, permission):
	"""
	Validates that the given request is made by an authorized user, identified either by API key or existing Flask
	session.

	Must be executed in an existing Flask request context!

	:param request: The Flask request object
	:param request: The required permission
	"""

	user = get_flask_user_from_request(request)
	if not user.has_permission(permission):
		raise tornado.web.HTTPError(403)

@deprecated("admin_validator is deprecated, please use new permission_validator", since="")
def admin_validator(request):
	from octoprint.access.permissions import Permissions
	return permission_validator(request, Permissions.ADMIN)

@deprecated("user_validator is deprecated, please use new permission_validator", since="")
def user_validator(request):
	return True

def get_flask_user_from_request(request):
	"""
	Retrieves the current flask user from the request context. Uses API key if available, otherwise the current
	user session if available.

	:param request: flask request from which to retrieve the current user
	:return: the user (might be an anonymous user)
	"""
	import octoprint.server.util
	import flask_login

	user = None

	apikey = octoprint.server.util.get_api_key(request)
	if apikey is not None:
		# user from api key?
		user = octoprint.server.util.get_user_for_apikey(apikey)

	if user is None:
		# user still None -> current session user
		user = flask_login.current_user

	if user is None:
		# user still None -> anonymous
		from octoprint.server import userManager
		user = userManager.anonymous_user_factory()

	return user


def redirect_to_tornado(request, target, code=302):
	"""
	Redirects from flask to tornado, flask request context must exist.

	:param request:
	:param target:
	:param code:
	:return:
	"""

	import flask

	requestUrl = request.url
	appBaseUrl = requestUrl[:requestUrl.find(flask.url_for("index") + "api")]

	redirectUrl = appBaseUrl + target
	if "?" in requestUrl:
		fragment = requestUrl[requestUrl.rfind("?"):]
		redirectUrl += fragment
	return flask.redirect(redirectUrl, code=code)


def restricted_access(func):
	"""
	This combines :py:func:`no_firstrun_access` and ``login_required``.
	"""
	@functools.wraps(func)
	def decorated_view(*args, **kwargs):
		return no_firstrun_access(flask_login.login_required(func))(*args, **kwargs)
	return decorated_view


def no_firstrun_access(func):
	"""
	If you decorate a view with this, it will ensure that first setup has been
	done for OctoPrint's Access Control.

	If OctoPrint's Access Control has not been setup yet (indicated by the "firstRun"
	flag from the settings being set to True and the userManager not indicating
	that it's user database has been customized from default), the decorator
	will cause a HTTP 403 status code to be returned by the decorated resource.
	"""
	@functools.wraps(func)
	def decorated_view(*args, **kwargs):
		# if OctoPrint hasn't been set up yet, abort
		if settings().getBoolean(["server", "firstRun"]) and settings().getBoolean(["accessControl", "enabled"]) and (octoprint.server.userManager is None or not octoprint.server.userManager.has_been_customized()):
			return flask.make_response("OctoPrint isn't setup yet", 403)
		return func(*args, **kwargs)

	return decorated_view


def firstrun_only_access(func):
	"""
	If you decorate a view with this, it will ensure that first setup has _not_ been
	done for OctoPrint's Access Control. Otherwise it
	will cause a HTTP 403 status code to be returned by the decorated resource.
	"""
	@functools.wraps(func)
	def decorated_view(*args, **kwargs):
		# if OctoPrint has been set up yet, abort
		if settings().getBoolean(["server", "firstRun"]) and (octoprint.server.userManager is None or not octoprint.server.userManager.has_been_customized()):
			return func(*args, **kwargs)
		else:
			return flask.make_response("OctoPrint is already setup, this resource is not longer available.", 403)

	return decorated_view


def get_remote_address(request):
	forwardedFor = request.headers.get("X-Forwarded-For", None)
	if forwardedFor is not None:
		return forwardedFor.split(",")[0]
	return request.remote_addr


def get_json_command_from_request(request, valid_commands):
	content_type = request.headers.get("Content-Type", None)
	if content_type is None or not "application/json" in content_type:
		return None, None, make_response("Expected content-type JSON", 400)

	data = request.get_json()
	if data is None:
		return None, None, make_response("Malformed JSON body or wrong content-type in request", 400)
	if not "command" in data or not data["command"] in valid_commands:
		return None, None, make_response("Expected valid command", 400)

	command = data["command"]
	for parameter in valid_commands[command]:
		if not parameter in data:
			return None, None, make_response("Mandatory parameter %s missing for command %s" % (parameter, command), 400)

	return command, data, None

##~~ Flask-Assets resolver with plugin asset support

class PluginAssetResolver(flask_assets.FlaskResolver):

	def split_prefix(self, ctx, item):
		app = ctx.environment._app
		if item.startswith("plugin/"):
			try:
				prefix, plugin, name = item.split("/", 2)
				blueprint = prefix + "." + plugin

				directory = flask_assets.get_static_folder(app.blueprints[blueprint])
				item = name
				endpoint = blueprint + ".static"
				return directory, item, endpoint
			except (ValueError, KeyError):
				pass

		return flask_assets.FlaskResolver.split_prefix(self, ctx, item)

	def resolve_output_to_path(self, ctx, target, bundle):
		import os
		return os.path.normpath(os.path.join(ctx.environment.directory, target))

##~~ Webassets updater that takes changes in the configuration into account

class SettingsCheckUpdater(webassets.updater.BaseUpdater):

	updater = "always"

	def __init__(self):
		self._delegate = webassets.updater.get_updater(self.__class__.updater)

	def needs_rebuild(self, bundle, ctx):
		return self._delegate.needs_rebuild(bundle, ctx) or self.changed_settings(ctx)

	def changed_settings(self, ctx):
		import json

		if not ctx.cache:
			return False

		cache_key = ('octo', 'settings')
		current_hash = webassets.utils.hash_func(json.dumps(settings().effective_yaml))
		cached_hash = ctx.cache.get(cache_key)
		# This may seem counter-intuitive, but if no cache entry is found
		# then we actually return "no update needed". This is because
		# otherwise if no cache / a dummy cache is used, then we would be
		# rebuilding every single time.
		if not cached_hash is None:
			return cached_hash != current_hash
		return False

	def build_done(self, bundle, ctx):
		import json

		self._delegate.build_done(bundle, ctx)
		if not ctx.cache:
			return

		cache_key = ('octo', 'settings')
		cache_value = webassets.utils.hash_func(json.dumps(settings().effective_yaml))
		ctx.cache.set(cache_key, cache_value)

##~~ core assets collector
def collect_core_assets(enable_gcodeviewer=True, preferred_stylesheet="css"):
	assets = dict(
		js=[],
		clientjs=[],
		css=[],
		less=[]
	)
	assets["js"] = [
		'js/app/bindings/allowbindings.js',
		'js/app/bindings/contextmenu.js',
		'js/app/bindings/copywidth.js',
		'js/app/bindings/invisible.js',
		'js/app/bindings/popover.js',
		'js/app/bindings/qrcode.js',
		'js/app/bindings/slimscrolledforeach.js',
		'js/app/bindings/toggle.js',
		'js/app/bindings/togglecontent.js',
		'js/app/bindings/valuewithinit.js',
		'js/app/viewmodels/access.js',
		'js/app/viewmodels/appearance.js',
		'js/app/viewmodels/connection.js',
		'js/app/viewmodels/control.js',
		'js/app/viewmodels/files.js',
		'js/app/viewmodels/loginstate.js',
		'js/app/viewmodels/navigation.js',
		'js/app/viewmodels/printerstate.js',
		'js/app/viewmodels/printerprofiles.js',
		'js/app/viewmodels/settings.js',
		'js/app/viewmodels/slicing.js',
		'js/app/viewmodels/system.js',
		'js/app/viewmodels/temperature.js',
		'js/app/viewmodels/terminal.js',
		'js/app/viewmodels/timelapse.js',
		'js/app/viewmodels/uistate.js',
		'js/app/viewmodels/users.js',
		'js/app/viewmodels/usersettings.js',
		'js/app/viewmodels/wizard.js',
		'js/app/viewmodels/about.js'
	]
	if enable_gcodeviewer:
		assets["js"] += [
			'js/app/viewmodels/gcode.js',
			'gcodeviewer/js/ui.js',
			'gcodeviewer/js/gCodeReader.js',
			'gcodeviewer/js/renderer.js'
		]

	assets["clientjs"] = [
		"js/app/client/base.js",
		"js/app/client/socket.js",
		"js/app/client/access.js",
		"js/app/client/browser.js",
		"js/app/client/connection.js",
		"js/app/client/control.js",
		"js/app/client/files.js",
		"js/app/client/job.js",
		"js/app/client/languages.js",
		"js/app/client/printer.js",
		"js/app/client/printerprofiles.js",
		"js/app/client/settings.js",
		"js/app/client/slicing.js",
		"js/app/client/system.js",
		"js/app/client/timelapse.js",
		"js/app/client/users.js",
		"js/app/client/util.js",
		"js/app/client/wizard.js"
	]

	if preferred_stylesheet == "less":
		assets["less"].append('less/octoprint.less')
	elif preferred_stylesheet == "css":
		assets["css"].append('css/octoprint.css')

	return assets

##~~ plugin assets collector

def collect_plugin_assets(enable_gcodeviewer=True, preferred_stylesheet="css"):
	logger = logging.getLogger(__name__ + ".collect_plugin_assets")

	supported_stylesheets = ("css", "less")
	assets = dict(bundled=dict(js=DefaultOrderedDict(list),
	                           clientjs=DefaultOrderedDict(list),
	                           css=DefaultOrderedDict(list),
	                           less=DefaultOrderedDict(list)),
	              external=dict(js=DefaultOrderedDict(list),
	                            clientjs=DefaultOrderedDict(list),
	                            css=DefaultOrderedDict(list),
	                            less=DefaultOrderedDict(list)))

	asset_plugins = octoprint.plugin.plugin_manager().get_implementations(octoprint.plugin.AssetPlugin)
	for implementation in asset_plugins:
		name = implementation._identifier
		is_bundled = implementation._plugin_info.bundled

		asset_key = "bundled" if is_bundled else "external"

		try:
			all_assets = implementation.get_assets()
			basefolder = implementation.get_asset_folder()
		except Exception:
			logger.exception("Got an error while trying to collect assets from {}, ignoring assets from the plugin".format(name),
			                 extra=dict(plugin=name))
			continue

		def asset_exists(category, asset):
			exists = os.path.exists(os.path.join(basefolder, asset))
			if not exists:
				logger.warning("Plugin {} is referring to non existing {} asset {}".format(name, category, asset))
			return exists

		if "js" in all_assets:
			for asset in all_assets["js"]:
				if not asset_exists("js", asset):
					continue
				assets[asset_key]["js"][name].append('plugin/{name}/{asset}'.format(**locals()))

		if "clientjs" in all_assets:
			for asset in all_assets["clientjs"]:
				if not asset_exists("clientjs", asset):
					continue
				assets[asset_key]["clientjs"][name].append("plugin/{name}/{asset}".format(**locals()))

		if preferred_stylesheet in all_assets:
			for asset in all_assets[preferred_stylesheet]:
				if not asset_exists(preferred_stylesheet, asset):
					continue
				assets[asset_key][preferred_stylesheet][name].append('plugin/{name}/{asset}'.format(**locals()))
		else:
			for stylesheet in supported_stylesheets:
				if not stylesheet in all_assets:
					continue

				for asset in all_assets[stylesheet]:
					if not asset_exists(stylesheet, asset):
						continue
					assets[asset_key][stylesheet][name].append('plugin/{name}/{asset}'.format(**locals()))
				break

	return assets

##~~ JSON encoding

class OctoPrintJsonEncoder(flask.json.JSONEncoder):
	def default(self, obj):
		try:
			return JsonEncoding.encode(obj)
		except TypeError:
			return flask.json.JSONEncoder.default(self, obj)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"

import octoprint.plugin

import sys
import inspect
from flask_babel import gettext
from octoprint.access import ADMIN_GROUP, USER_GROUP
from octoprint.util import to_native_str

# noinspection PyUnresolvedReferences,PyMethodMayBeStatic
class ServerCommandsSubwizard(object):
	def _is_servercommands_wizard_firstrunonly(self):
		return True

	def _is_servercommands_wizard_required(self):
		system_shutdown_command = self._settings.global_get(["server", "commands", "systemShutdownCommand"])
		system_restart_command = self._settings.global_get(["server", "commands", "systemRestartCommand"])
		server_restart_command = self._settings.global_get(["server", "commands", "serverRestartCommand"])

		return not (system_shutdown_command and system_restart_command and server_restart_command)

	def _get_servercommands_wizard_details(self):
		return dict(required=self._is_servercommands_wizard_required())

	def _get_servercommands_wizard_name(self):
		return gettext("Server Commands")


# noinspection PyUnresolvedReferences,PyMethodMayBeStatic
class WebcamSubwizard(object):
	def _is_webcam_wizard_firstrunonly(self):
		return True

	def _is_webcam_wizard_required(self):
		webcam_snapshot_url = self._settings.global_get(["webcam", "snapshot"])
		webcam_stream_url = self._settings.global_get(["webcam", "stream"])
		ffmpeg_path = self._settings.global_get(["webcam", "ffmpeg"])

		return not (webcam_snapshot_url and webcam_stream_url and ffmpeg_path)

	def _get_webcam_wizard_details(self):
		return dict(required=self._is_webcam_wizard_required())

	def _get_webcam_wizard_name(self):
		return gettext("Webcam & Timelapse")


# noinspection PyUnresolvedReferences,PyMethodMayBeStatic
class AclSubwizard(object):
	def _is_acl_wizard_firstrunonly(self):
		return True

	def _is_acl_wizard_required(self):
		return self._user_manager.enabled and not self._user_manager.has_been_customized()

	def _get_acl_wizard_details(self):
		return dict(required=self._is_acl_wizard_required())

	def _get_acl_wizard_name(self):
		return gettext("Access Control")

	def _get_acl_additional_wizard_template_data(self):
		return dict(mandatory=self._is_acl_wizard_required())

	@octoprint.plugin.BlueprintPlugin.route("/acl", methods=["POST"])
	def acl_wizard_api(self):
		from flask import request, abort
		from octoprint.server.api import valid_boolean_trues, NO_CONTENT

		if not self._settings.global_get(["server", "firstRun"]) or self._user_manager.has_been_customized():
			abort(404)

		data = request.get_json()
		if data is None:
			data = request.values

		if "ac" in data and data["ac"] in valid_boolean_trues and \
						"user" in data and "pass1" in data and \
						"pass2" in data and data["pass1"] == data["pass2"]:
			# configure access control
			self._settings.global_set_boolean(["accessControl", "enabled"], True)
			self._user_manager.enable()

			self._user_manager.add_user(data["user"], data["pass1"], True, [], [USER_GROUP, ADMIN_GROUP], overwrite=True)
		elif "ac" in data and not data["ac"] in valid_boolean_trues:
			# disable access control
			self._settings.global_set_boolean(["accessControl", "enabled"], False)
			self._user_manager.disable()
		self._settings.save()
		return NO_CONTENT


# noinspection PyUnresolvedReferences,PyMethodMayBeStatic
class OnlineCheckSubwizard(object):
	def _is_onlinecheck_wizard_firstrunonly(self):
		return False

	def _is_onlinecheck_wizard_required(self):
		return self._settings.global_get(["server", "onlineCheck", "enabled"]) is None

	def _get_onlinecheck_wizard_details(self):
		return dict(required=self._is_onlinecheck_wizard_required())

	def _get_onlinecheck_wizard_name(self):
		return gettext("Online connectivity check")

	def _get_onlinecheck_additional_wizard_template_data(self):
		return dict(mandatory=self._is_onlinecheck_wizard_required())


# noinspection PyUnresolvedReferences,PyMethodMayBeStatic
class PluginBlacklistSubwizard(object):
	def _is_pluginblacklist_wizard_firstrunonly(self):
		return False

	def _is_pluginblacklist_wizard_required(self):
		return self._settings.global_get(["server", "pluginBlacklist", "enabled"]) is None

	def _get_pluginblacklist_wizard_details(self):
		return dict(required=self._is_pluginblacklist_wizard_required())

	def _get_pluginblacklist_wizard_name(self):
		return gettext("Plugin blacklist")

	def _get_pluginblacklist_additional_wizard_template_data(self):
		return dict(mandatory=self._is_pluginblacklist_wizard_required())


# noinspection PyUnresolvedReferences,PyMethodMayBeStatic
class PrinterProfileSubwizard(object):
	def _is_printerprofile_wizard_firstrunonly(self):
		return True

	def _is_printerprofile_wizard_required(self):
		return self._printer_profile_manager.is_default_unmodified() and self._printer_profile_manager.profile_count == 1

	def _get_printerprofile_wizard_details(self):
		return dict(required=self._is_printerprofile_wizard_required())

	def _get_printerprofile_wizard_name(self):
		return gettext("Default Printer Profile")


Subwizards = type(to_native_str("Subwizards"),
                  tuple(cls for clsname, cls in inspect.getmembers(sys.modules[__name__], inspect.isclass)
                        if clsname.endswith("Subwizard")),
                  dict())



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

"""
Slicing related exceptions.

.. autoclass:: SlicingException

.. autoclass:: SlicingCancelled
   :show-inheritance:

.. autoclass:: SlicerException
   :show-inheritance:

.. autoclass:: UnknownSlicer
   :show-inheritance:

.. autoclass:: SlicerNotConfigured
   :show-inheritance:

.. autoclass:: ProfileException

.. autoclass:: UnknownProfile
   :show-inheritance:

.. autoclass:: ProfileAlreadyExists
   :show-inheritance:

"""

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"


class SlicingException(Exception):
	"""
	Base exception of all slicing related exceptions.
	"""
	pass

class SlicingCancelled(SlicingException):
	"""
	Raised if a slicing job was cancelled.
	"""
	pass

class SlicerException(SlicingException):
	"""
	Base exception of all slicer related exceptions.

	.. attribute:: slicer

	   Identifier of the slicer for which the exception was raised.
	"""
	def __init__(self, slicer, *args, **kwargs):
		SlicingException.__init__(self, *args, **kwargs)
		self.slicer = slicer

class SlicerNotConfigured(SlicerException):
	"""
	Raised if a slicer is not yet configured but must be configured to proceed.
	"""
	def __init__(self, slicer, *args, **kwargs):
		SlicerException.__init__(self, slicer, *args, **kwargs)
		self.message = "Slicer not configured: {slicer}".format(slicer=slicer)

class UnknownSlicer(SlicerException):
	"""
	Raised if a slicer is unknown.
	"""
	def __init__(self, slicer, *args, **kwargs):
		SlicerException.__init__(self, slicer, *args, **kwargs)
		self.message = "No such slicer: {slicer}".format(slicer=slicer)

class ProfileException(Exception):
	"""
	Base exception of all slicing profile related exceptions.

	.. attribute:: slicer

	   Identifier of the slicer to which the profile belongs.

	.. attribute:: profile

	   Identifier of the profile for which the exception was raised.
	"""
	def __init__(self, slicer, profile, *args, **kwargs):
		Exception.__init__(self, *args, **kwargs)
		self.slicer = slicer
		self.profile = profile

class UnknownProfile(ProfileException):
	"""
	Raised if a slicing profile does not exist but must exist to proceed.
	"""
	def __init__(self, slicer, profile, *args, **kwargs):
		ProfileException.__init__(self, slicer, profile, *args, **kwargs)
		self.message = "Profile {profile} for slicer {slicer} does not exist".format(profile=profile, slicer=slicer)

class ProfileAlreadyExists(ProfileException):
	"""
	Raised if a slicing profile already exists and must not be overwritten.
	"""
	def __init__(self, slicer, profile, *args, **kwargs):
		ProfileException.__init__(self, slicer, profile, *args, **kwargs)
		self.message = "Profile {profile} for slicer {slicer} already exists".format(profile=profile, slicer=slicer)

class CouldNotDeleteProfile(ProfileException):
	"""
	Raised if there is an unexpected error trying to delete a known profile.
	"""
	def __init__(self, slicer, profile, cause=None, *args, **kwargs):
		ProfileException.__init__(self, slicer, profile, *args, **kwargs)

		self.cause = cause
		if cause:
			self.message = "Could not delete profile {profile} for slicer {slicer}: {cause}".format(profile=profile, slicer=slicer, cause=str(cause))
		else:
			self.message = "Could not delete profile {profile} for slicer {slicer}".format(profile=profile, slicer=slicer)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from past.builtins import basestring, unicode

from octoprint.logging import handlers

from typing import Union

def log_to_handler(logger, handler, level, msg, exc_info=None, extra=None, *args):
	"""
	Logs to the provided handler only.

	Arguments:
		logger: logger to log to
		handler: handler to restrict logging to
		level: level to log at
		msg: message to log
		exc_info: optional exception info
		extra: optional extra data
		*args: log args
	"""
	from logging import _srcfile
	import sys

	# this is just the same as logging.Logger._log

	if _srcfile:
		# IronPython doesn't track Python frames, so findCaller raises an
		# exception on some versions of IronPython. We trap it here so that
		# IronPython can use logging.
		try:
			fn, lno, func = logger.findCaller()
		except ValueError:
			fn, lno, func = "(unknown file)", 0, "(unknown function)"
	else:
		fn, lno, func = "(unknown file)", 0, "(unknown function)"
	if exc_info:
		if not isinstance(exc_info, tuple):
			exc_info = sys.exc_info()

	record = logger.makeRecord(logger.name, level, fn, lno, msg, args, exc_info, func, extra)

	# and this is a mixture of logging.Logger.handle and logging.Logger.callHandlers

	if (not logger.disabled) and logger.filter(record):
		if record.levelno >= handler.level:
			handler.handle(record)


def get_handler(name, logger=None):
	"""
	Retrieves the handler named ``name``.

	If optional ``logger`` is provided, search will be
	limited to that logger, otherwise the root logger will be
	searched.

	Arguments:
		name: the name of the handler to look for
		logger: (optional) the logger to search in, root logger if not provided

	Returns:
	    the handler if it could be found, None otherwise
	"""
	import logging

	if logger is None:
		logger = logging.getLogger()

	for handler in logger.handlers:
		if handler.get_name() == name:
			return handler

	return None


def get_divider_line(c, message=None, length=78, indent=3):
	"""
	Generate a divider line for logging, optionally with included message.

	Examples:

	    >>> get_divider_line("-") # doctest: +ALLOW_UNICODE
	    '------------------------------------------------------------------------------'
	    >>> get_divider_line("=", length=10) # doctest: +ALLOW_UNICODE
	    '=========='
	    >>> get_divider_line("-", message="Hi", length=10) # doctest: +ALLOW_UNICODE
	    '--- Hi ---'
	    >>> get_divider_line("-", message="A slightly longer text") # doctest: +ALLOW_UNICODE
	    '--- A slightly longer text ---------------------------------------------------'
	    >>> get_divider_line("-", message="A slightly longer text", indent=5) # doctest: +ALLOW_UNICODE
	    '----- A slightly longer text -------------------------------------------------'
	    >>> get_divider_line("-", message="Hello World!", length=10) # doctest: +ALLOW_UNICODE
	    '--- Hello World!'
	    >>> get_divider_line(None)
	    Traceback (most recent call last):
	      ...
	    AssertionError: c is not text
	    >>> get_divider_line("´`")
	    Traceback (most recent call last):
	      ...
	    AssertionError: c is not a single character
	    >>> get_divider_line("-", message=3)
	    Traceback (most recent call last):
	      ...
	    AssertionError: message is not text
	    >>> get_divider_line("-", length="hello")
	    Traceback (most recent call last):
	      ...
	    AssertionError: length is not an int
	    >>> get_divider_line("-", indent="hi")
	    Traceback (most recent call last):
	      ...
	    AssertionError: indent is not an int

	Arguments:
		c: character to use for the line
		message: message to print in the line
		length: length of the line
		indent: indentation of message in line

	Returns:
		formatted divider line
	"""

	assert isinstance(c, basestring), "c is not text"
	assert len(c) == 1, "c is not a single character"
	assert isinstance(length, int), "length is not an int"
	assert isinstance(indent, int), "indent is not an int"

	if message is None:
		return c * length

	assert isinstance(message, basestring), "message is not text"

	space = length - 2 * (indent + 1)
	if space >= len(message):
		return c * indent + " " + message + " " + c * (length - indent - 2 - len(message))
	else:
		return c * indent + " " + message


def prefix_multilines(text, prefix=u": "):
	# type: (Union[unicode, bytes], unicode) -> unicode
	from octoprint.util import to_unicode

	lines = text.splitlines()
	if not lines:
		return u""

	if len(lines) == 1:
		return to_unicode(lines[0])

	return to_unicode(lines[0]) + u"\n" + u"\n".join(map(lambda line: prefix + to_unicode(line),
	                                                     lines[1:]))



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2017 The OctoPrint Project - Released under terms of the AGPLv3 License"

import copy
import logging
import os
import sys
import threading
import yaml

import psutil

from octoprint.plugin import EnvironmentDetectionPlugin
from octoprint.util.platform import get_os
from octoprint.util.version import get_python_version_string

class EnvironmentDetector(object):

	def __init__(self, plugin_manager):
		self._plugin_manager = plugin_manager

		self._cache = None
		self._cache_lock = threading.RLock()

		self._logger = logging.getLogger(__name__)

		try:
			self._environment_plugins = self._plugin_manager.get_implementations(EnvironmentDetectionPlugin)
		except Exception:
			# just in case, see #3100...
			self._logger.exception("There was an error fetching EnvironmentDetectionPlugins from the plugin manager")
			self._environment_plugins = []

	@property
	def environment(self):
		with self._cache_lock:
			if self._cache is None:
				self.run_detection()
			return copy.deepcopy(self._cache)

	def run_detection(self, notify_plugins=True):
		try:
			environment = dict()
			environment["os"] = self._detect_os()
			environment["python"] = self._detect_python()
			environment["hardware"] = self._detect_hardware()

			plugin_result = self._detect_from_plugins()
			if plugin_result:
				environment["plugins"] = plugin_result

			with self._cache_lock:
				self._cache = environment

			if notify_plugins:
				self.notify_plugins()

			return environment
		except Exception:
			self._logger.exception("Unexpected error while detecting environment")
			with self._cache_lock:
				self._cache = dict()
				return self._cache

	def _detect_os(self):
		return dict(id=get_os(),
		            platform=sys.platform)

	def _detect_python(self):
		result = dict(version="unknown",
		              pip="unknown")

		# determine python version
		try:
			result["version"] = get_python_version_string()
		except Exception:
			self._logger.exception("Error detecting python version")

		# determine if we are running from a virtual environment
		try:
			if hasattr(sys, "real_prefix") or (hasattr(sys, "base_prefix") and os.path.realpath(sys.prefix) != os.path.realpath(sys.base_prefix)):
				result["virtualenv"] = sys.prefix
		except Exception:
			self._logger.exception("Error detecting whether we are running in a virtual environment")

		# try to find pip version
		try:
			import pkg_resources
			result["pip"] = pkg_resources.get_distribution("pip").version
		except Exception:
			self._logger.exception("Error detecting pip version")

		return result

	def _detect_hardware(self):
		result = dict(cores="unknown",
		              freq="unknown",
		              ram="unknown")

		try:
			cores = psutil.cpu_count()
			cpu_freq = psutil.cpu_freq()
			ram = psutil.virtual_memory()
			if cores:
				result["cores"] = cores
			if cpu_freq and hasattr(cpu_freq, "max"):
				result["freq"] = cpu_freq.max
			if ram and hasattr(ram, "total"):
				result["ram"] = ram.total
		except Exception:
			self._logger.exception("Error while detecting hardware environment")

		return result

	def _detect_from_plugins(self):
		result = dict()

		for implementation in self._environment_plugins:
			try:
				additional = implementation.get_additional_environment()
				if additional is not None and isinstance(additional, dict) and len(additional):
					result[implementation._identifier] = additional
			except Exception:
				self._logger.exception("Error while fetching additional "
				                       "environment data from plugin {}".format(implementation._identifier),
				                       extra=dict(plugin=implementation._identifier))

		return result

	def log_detected_environment(self, only_to_handler=None):
		def _log(message, level=logging.INFO):
			if only_to_handler is not None:
				import octoprint.logging
				octoprint.logging.log_to_handler(self._logger, only_to_handler, level, message, [])
			else:
				self._logger.log(level, message)

		try:
			_log(self._format())
		except Exception:
			self._logger.exception("Error logging detected environment")

	def _format(self):
		with self._cache_lock:
			if self._cache is None:
				self.run_detection()
			environment = copy.deepcopy(self._cache)

		dumped_environment = yaml.safe_dump(environment,
		                                    default_flow_style=False,
		                                    indent=4,
		                                    allow_unicode=True).strip()
		environment_lines = "\n".join(map(lambda l: "|  {}".format(l), dumped_environment.split("\n")))
		return "Detected environment is Python {} under {} ({}). Details:\n{}".format(environment["python"]["version"],
		                                                                               environment["os"]["id"].title(),
		                                                                               environment["os"]["platform"],
		                                                                               environment_lines)

	def notify_plugins(self):
		with self._cache_lock:
			if self._cache is None:
				self.run_detection(notify_plugins=False)
			environment = copy.deepcopy(self._cache)

		for implementation in self._environment_plugins:
			try:
				implementation.on_environment_detected(environment)
			except Exception:
				self._logger.exception("Error while sending environment "
				                       "detection result to plugin {}".format(implementation._identifier))



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"


import octoprint.plugin

import io
import copy
import flask
import os
import threading
import time
import logging
import logging.handlers
import hashlib

# noinspection PyCompatibility
from concurrent import futures

from . import version_checks, updaters, exceptions, util, cli

from flask_babel import gettext

from octoprint.server.util.flask import no_firstrun_access, with_revalidation_checking, check_etag
from octoprint.server import VERSION, REVISION, BRANCH
from octoprint.access import USER_GROUP, ADMIN_GROUP
from octoprint.access.permissions import Permissions
from octoprint.util import dict_merge, to_unicode
from octoprint.util.version import get_comparable_version, get_python_version_string
from octoprint.util.pip import LocalPipCaller
import octoprint.settings


# OctoPi 0.15+
MINIMUM_PYTHON = "2.7.9"
MINIMUM_SETUPTOOLS = "39.0.1"
MINIMUM_PIP = "9.0.3"


##~~ Plugin

class SoftwareUpdatePlugin(octoprint.plugin.BlueprintPlugin,
                           octoprint.plugin.SettingsPlugin,
                           octoprint.plugin.AssetPlugin,
                           octoprint.plugin.TemplatePlugin,
                           octoprint.plugin.StartupPlugin,
                           octoprint.plugin.WizardPlugin,
                           octoprint.plugin.EventHandlerPlugin):

	COMMIT_TRACKING_TYPES = ("github_commit", "bitbucket_commit")

	DATA_FORMAT_VERSION = "v4"

	# noinspection PyMissingConstructor
	def __init__(self):
		self._update_in_progress = False
		self._configured_checks_mutex = threading.Lock()
		self._configured_checks = None
		self._refresh_configured_checks = False

		self._get_versions_mutex = threading.RLock()
		self._get_versions_data = None
		self._get_versions_data_ready = threading.Event()

		self._version_cache = dict()
		self._version_cache_ttl = 0
		self._version_cache_path = None
		self._version_cache_dirty = False
		self._version_cache_timestamp = None

		self._environment_supported = True
		self._environment_versions = dict()
		self._environment_ready = threading.Event()

		self._console_logger = None

		self._get_throttled = lambda: False

	def initialize(self):
		self._console_logger = logging.getLogger("octoprint.plugins.softwareupdate.console")

		self._version_cache_ttl = self._settings.get_int(["cache_ttl"]) * 60
		self._version_cache_path = os.path.join(self.get_plugin_data_folder(), "versioncache.yaml")
		self._load_version_cache()

		def refresh_checks(name, plugin):
			self._refresh_configured_checks = True
			self._send_client_message("update_versions")

		self._plugin_lifecycle_manager.add_callback("enabled", refresh_checks)
		self._plugin_lifecycle_manager.add_callback("disabled", refresh_checks)

	# Additional permissions hook

	def get_additional_permissions(self):
		return [
			dict(key="CHECK",
			     name="Check",
			     description=gettext("Allows to check for software updates"),
			     roles=["check"],
			     default_groups=[USER_GROUP]),
			dict(key="UPDATE",
			     name="Update",
			     description=gettext("Allows to perform software updates"),
			     default_groups=[ADMIN_GROUP],
			     roles=["update"],
			     dangerous=True)
		]

	def on_startup(self, host, port):
		console_logging_handler = logging.handlers.RotatingFileHandler(self._settings.get_plugin_logfile_path(postfix="console"),
		                                                               maxBytes=2*1024*1024,
		                                                               encoding="utf-8")
		console_logging_handler.setFormatter(logging.Formatter("%(asctime)s %(message)s"))
		console_logging_handler.setLevel(logging.DEBUG)

		self._console_logger.addHandler(console_logging_handler)
		self._console_logger.setLevel(logging.DEBUG)
		self._console_logger.propagate = False

		helpers = self._plugin_manager.get_helpers("pi_support", "get_throttled")
		if helpers and "get_throttled" in helpers:
			self._get_throttled = helpers["get_throttled"]
			if self._settings.get_boolean(["ignore_throttled"]):
				self._logger.warn("!!! THROTTLE STATE IGNORED !!! You have configured the Software Update plugin to ignore an active throttle state of the underlying system. You might run into stability issues or outright corrupt your install. Consider fixing the throttling issue instead of suppressing it.")

	def on_after_startup(self):
		self._check_environment()
		self.get_current_versions()

	def _get_configured_checks(self):
		with self._configured_checks_mutex:
			if self._refresh_configured_checks or self._configured_checks is None:
				self._refresh_configured_checks = False
				self._configured_checks = self._settings.get(["checks"], merged=True)

				update_check_hooks = self._plugin_manager.get_hooks("octoprint.plugin.softwareupdate.check_config")
				check_providers = dict()

				effective_configs = dict()

				for name, hook in update_check_hooks.items():
					try:
						hook_checks = hook()
					except Exception:
						self._logger.exception("Error while retrieving update information "
						                       "from plugin {name}".format(**locals()),
						                       extra=dict(plugin=name))
					else:
						for key, default_config in hook_checks.items():
							if key in effective_configs or key == "octoprint":
								if key == name:
									self._logger.warning("Software update hook {} provides check for itself but that was already registered by {} - overwriting that third party registration now!".format(name, check_providers.get(key, "unknown hook")))
								else:
									self._logger.warning("Software update hook {} tried to overwrite config for check {} but that was already configured elsewhere".format(name, key))
									continue

							check_providers[key] = name

							yaml_config = dict()
							effective_config = default_config
							if key in self._configured_checks:
								yaml_config = self._configured_checks[key]
								effective_config = dict_merge(default_config, yaml_config)

								# Make sure there's nothing persisted in that check that shouldn't be persisted
								#
								# This used to be part of the settings migration (version 2) due to a bug - it can't
								# stay there though since it interferes with manual entries to the checks not
								# originating from within a plugin. Hence we do that step now here.
								if "type" not in effective_config or effective_config["type"] not in self.COMMIT_TRACKING_TYPES:
									deletables = ["current", "displayVersion"]
								else:
									deletables = []
								self._clean_settings_check(key, yaml_config, default_config, delete=deletables, save=False)

							if effective_config:
								effective_configs[key] = effective_config
							else:
								self._logger.warning("Update for {} is empty or None, ignoring it".format(key))

				# finally set all our internal representations to our processed results
				for key, config in effective_configs.items():
					self._configured_checks[key] = config

				# we only want to process checks that came from plugins for
				# which the plugins are still installed and enabled
				config_checks = self._settings.get(["checks"])
				plugin_and_not_enabled = lambda k: k in check_providers and \
				                                   not check_providers[k] in self._plugin_manager.enabled_plugins
				obsolete_plugin_checks = list(filter(plugin_and_not_enabled,
				                                	 config_checks.keys()))
				for key in obsolete_plugin_checks:
					self._logger.debug("Check for key {} was provided by plugin {} that's no longer available, ignoring it".format(key, check_providers[key]))
					del self._configured_checks[key]

			return self._configured_checks

	def _check_environment(self):
		import pkg_resources

		local_pip = LocalPipCaller()

		# check python and setuptools version
		versions = dict(python=get_python_version_string(),
		                setuptools=pkg_resources.get_distribution("setuptools").version,
		                pip=local_pip.version_string)
		supported = get_comparable_version(versions["python"]) >= get_comparable_version(MINIMUM_PYTHON) \
		       and get_comparable_version(versions["setuptools"]) >= get_comparable_version(MINIMUM_SETUPTOOLS) \
		       and get_comparable_version(versions["pip"]) >= get_comparable_version(MINIMUM_PIP)

		self._environment_supported = supported
		self._environment_versions = versions
		self._environment_ready.set()

	def _load_version_cache(self):
		if not os.path.isfile(self._version_cache_path):
			return

		import yaml
		try:
			with io.open(self._version_cache_path, 'rt', encoding='utf-8') as f:
				data = yaml.safe_load(f)
			timestamp = os.stat(self._version_cache_path).st_mtime
		except Exception:
			self._logger.exception("Error while loading version cache from disk")
		else:
			try:
				if not isinstance(data, dict):
					self._logger.info("Version cache was created in a different format, not using it")
					return

				if "__version" in data:
					data_version = data["__version"]
				else:
					self._logger.info("Can't determine version of OctoPrint version cache was created for, not using it")
					return

				from octoprint._version import get_versions
				octoprint_version = get_versions()["version"]
				if data_version != octoprint_version:
					self._logger.info("Version cache was created for another version of OctoPrint, not using it")
					return

				self._version_cache = data
				self._version_cache_dirty = False
				self._version_cache_timestamp = timestamp
				self._logger.info("Loaded version cache from disk")
			except Exception:
				self._logger.exception("Error parsing in version cache data")

	def _save_version_cache(self):
		import yaml
		from octoprint.util import atomic_write
		from octoprint._version import get_versions

		octoprint_version = get_versions()["version"]
		self._version_cache["__version"] = octoprint_version

		with atomic_write(self._version_cache_path, mode='wt', max_permissions=0o666) as file_obj:
			yaml.safe_dump(self._version_cache, stream=file_obj, default_flow_style=False, indent=2, allow_unicode=True)

		self._version_cache_dirty = False
		self._version_cache_timestamp = time.time()
		self._logger.info("Saved version cache to disk")

	#~~ SettingsPlugin API

	def get_settings_defaults(self):
		update_script = os.path.join(self._basefolder, "scripts", "update-octoprint.py")
		default_update_script = "{{python}} \"{update_script}\" --branch={{branch}} " \
		                        "--force={{force}} \"{{folder}}\" {{target}}".format(update_script=update_script)

		return {
			"checks": {
				"octoprint": {
					"type": "github_release",
					"user": "foosel",
					"repo": "OctoPrint",
					"method": "pip",
					"pip": "https://github.com/foosel/OctoPrint/archive/{target_version}.zip",
					"update_script": default_update_script,
					"restart": "octoprint",
					"stable_branch": dict(branch="master", commitish=["master"], name="Stable"),
					"prerelease_branches": [dict(branch="rc/maintenance",
					                             commitish=["rc/maintenance"],             # maintenance RCs
					                             name="Maintenance RCs"),
					                        dict(branch="rc/devel",
					                             commitish=["rc/maintenance", "rc/devel"], # devel & maintenance RCs
					                             name="Devel RCs")]
				},
			},
			"pip_command": None,

			"cache_ttl": 24 * 60,

			"notify_users": True,

			"ignore_throttled": False
		}

	def on_settings_load(self):
		data = dict(octoprint.plugin.SettingsPlugin.on_settings_load(self))
		if "checks" in data:
			del data["checks"]

		checks = self._get_configured_checks()
		if "octoprint" in checks:
			data["octoprint_checkout_folder"] = self._get_octoprint_checkout_folder(checks=checks)
			data["octoprint_tracked_branch"] = self._get_octoprint_tracked_branch(checks=checks)
			data["octoprint_pip_target"] = self._get_octoprint_pip_target(checks=checks)
			data["octoprint_type"] = checks["octoprint"].get("type", None)

			try:
				data["octoprint_method"] = self._get_update_method("octoprint", checks["octoprint"])
			except exceptions.UnknownUpdateType:
				data["octoprint_method"] = "unknown"

			stable_branch = None
			prerelease_branches = []
			branch_mappings = []
			if "stable_branch" in checks["octoprint"]:
				branch_mappings.append(checks["octoprint"]["stable_branch"])
				stable_branch = checks["octoprint"]["stable_branch"]["branch"]
			if "prerelease_branches" in checks["octoprint"]:
				for mapping in checks["octoprint"]["prerelease_branches"]:
					branch_mappings.append(mapping)
					prerelease_branches.append(mapping["branch"])
			data["octoprint_branch_mappings"] = branch_mappings

			data["octoprint_release_channel"] = stable_branch
			if checks["octoprint"].get("prerelease", False):
				channel = checks["octoprint"].get("prerelease_channel", BRANCH)
				if channel in prerelease_branches:
					data["octoprint_release_channel"] = channel

		else:
			data["octoprint_checkout_folder"] = None
			data["octoprint_type"] = None
			data["octoprint_branch_mappings"] = []

		data["pip_enable_check"] = "pip" in checks

		return data

	def on_settings_save(self, data):
		for key in self.get_settings_defaults():
			if key in ("checks", "cache_ttl", "notify_user", "octoprint_checkout_folder", "octoprint_type", "octoprint_release_channel"):
				continue
			if key in data:
				self._settings.set([key], data[key])

		if "cache_ttl" in data:
			self._settings.set_int(["cache_ttl"], data["cache_ttl"])
		self._version_cache_ttl = self._settings.get_int(["cache_ttl"]) * 60

		if "notify_users" in data:
			self._settings.set_boolean(["notify_users"], data["notify_users"])

		defaults = dict(
			plugins=dict(softwareupdate=dict(
				checks=dict(
					octoprint=self.get_settings_defaults()["checks"]["octoprint"]
				)
			))
		)

		updated_octoprint_check_config = False
		update_pip_check_config = False

		if "octoprint_checkout_folder" in data:
			self._settings.set(["checks", "octoprint", "checkout_folder"], data["octoprint_checkout_folder"], defaults=defaults, force=True)
			if self._settings.get(["checks", "octoprint", "update_folder"]) and data["octoprint_checkout_folder"]:
				self._settings.set(["checks", "octoprint", "update_folder"], None, defaults=defaults, force=True)
			updated_octoprint_check_config = True

		if "octoprint_type" in data:
			octoprint_type = data["octoprint_type"]

			if octoprint_type == "github_release":
				self._settings.set(["checks", "octoprint", "type"], octoprint_type, defaults=defaults)
				self._settings.set(["checks", "octoprint", "method"], "pip", defaults=defaults)
				updated_octoprint_check_config = True

			elif octoprint_type == "github_commit":
				self._settings.set(["checks", "octoprint", "type"], octoprint_type, defaults=defaults)
				self._settings.set(["checks", "octoprint", "method"], "pip", defaults=defaults)
				updated_octoprint_check_config = True

			elif octoprint_type == "git_commit":
				self._settings.set(["checks", "octoprint", "type"], octoprint_type, defaults=defaults)
				self._settings.set(["checks", "octoprint", "method"], "update_script", defaults=defaults)
				updated_octoprint_check_config = True

		if "octoprint_tracked_branch" in data:
			self._settings.set(["checks", "octoprint", "branch"], data["octoprint_tracked_branch"], defaults=defaults, force=True)
			updated_octoprint_check_config = True

		if "octoprint_pip_target" in data:
			self._settings.set(["checks", "octoprint", "pip"], data["octoprint_pip_target"], defaults=defaults)
			updated_octoprint_check_config = True

		if "octoprint_release_channel" in data:
			prerelease_branches = self._settings.get(["checks", "octoprint", "prerelease_branches"])
			if prerelease_branches and data["octoprint_release_channel"] in [x["branch"] for x in prerelease_branches]:
				self._settings.set(["checks", "octoprint", "prerelease"], True, defaults=defaults, force=True)
				self._settings.set(["checks", "octoprint", "prerelease_channel"], data["octoprint_release_channel"],
				                   defaults=defaults, force=True)
			else:
				self._settings.set(["checks", "octoprint", "prerelease"], False, defaults=defaults, force=True)
				self._settings.set(["checks", "octoprint", "prerelease_channel"], None, defaults=defaults, force=True)
			updated_octoprint_check_config = True

		if "pip_enable_check" in data:
			checks = self._settings.get(["checks"], merged=True)
			if data["pip_enable_check"] in octoprint.settings.valid_boolean_trues:
				checks["pip"] = dict(type="pypi_release",
				                     package="pip",
				                     pip="pip=={target_version}")
			elif "pip" in checks:
				del checks["pip"]
			self._settings.set(["checks"], checks, force=True)
			self._settings.save()
			update_pip_check_config = True

		if updated_octoprint_check_config:
			self._refresh_configured_checks = True
			try:
				del self._version_cache["octoprint"]
			except KeyError:
				pass
			self._version_cache_dirty = True

		if update_pip_check_config:
			self._refresh_configured_checks = True
			try:
				del self._version_cache["pip"]
			except KeyError:
				pass
			self._version_cache_dirty = True

	def get_settings_version(self):
		return 7

	def on_settings_migrate(self, target, current=None):

		if current is None or current < 6:
			# up until & including config version 5 we didn't set the method parameter for the octoprint check
			# configuration

			configured_checks = self._settings.get(["checks"], incl_defaults=False)
			if configured_checks is not None and "octoprint" in configured_checks:
				octoprint_check = dict(configured_checks["octoprint"])

				if not "method" in octoprint_check and octoprint_check.get("type") == "git_commit":
					defaults = dict(plugins=dict(softwareupdate=dict(checks=dict(octoprint=dict(method="pip")))))
					self._settings.set(["checks", "octoprint", "method"], "update_script", defaults=defaults)

		if current == 4:
			# config version 4 didn't correctly remove the old settings for octoprint_restart_command
			# and environment_restart_command

			self._settings.set(["environment_restart_command"], None)
			self._settings.set(["octoprint_restart_command"], None)

		if current is None or current < 5:
			# config version 4 and higher moves octoprint_restart_command and
			# environment_restart_command to the core configuration

			# current plugin commands
			configured_octoprint_restart_command = self._settings.get(["octoprint_restart_command"])
			configured_environment_restart_command = self._settings.get(["environment_restart_command"])

			# current global commands
			configured_system_restart_command = self._settings.global_get(["server", "commands", "systemRestartCommand"])
			configured_server_restart_command = self._settings.global_get(["server", "commands", "serverRestartCommand"])

			# only set global commands if they are not yet set
			if configured_system_restart_command is None and configured_environment_restart_command is not None:
				self._settings.global_set(["server", "commands", "systemRestartCommand"], configured_environment_restart_command)
			if configured_server_restart_command is None and configured_octoprint_restart_command is not None:
				self._settings.global_set(["server", "commands", "serverRestartCommand"], configured_octoprint_restart_command)

			# delete current plugin commands from config
			self._settings.set(["environment_restart_command"], None)
			self._settings.set(["octoprint_restart_command"], None)

		if current is None or current == 2:
			# No config version and config version 2 need the same fix, stripping
			# accidentally persisted data off the checks.
			#
			# We used to do the same processing for the plugin entries too here, but that interfered
			# with manual configuration entries. Stuff got deleted that wasn't supposed to be deleted.
			#
			# The problem is that we don't know if an entry we are looking at and which didn't come through
			# a plugin hook is simply an entry from a now uninstalled/unactive plugin, or if it was something
			# manually configured by the user. So instead of just blindly removing anything that doesn't
			# come from a plugin here we instead clean up anything that indeed comes from a plugin
			# during run time and leave everything else as is in the hopes that will not cause trouble.
			#
			# We still handle the "octoprint" entry here though.

			configured_checks = self._settings.get(["checks"], incl_defaults=False)
			if configured_checks is not None and "octoprint" in configured_checks:
				octoprint_check = dict(configured_checks["octoprint"])
				if "type" not in octoprint_check or octoprint_check["type"] not in self.COMMIT_TRACKING_TYPES:
					deletables=["current", "displayName", "displayVersion"]
				else:
					deletables=[]
				self._clean_settings_check("octoprint", octoprint_check, self.get_settings_defaults()["checks"]["octoprint"], delete=deletables, save=False)

		elif current == 1:
			# config version 1 had the error that the octoprint check got accidentally
			# included in checks["octoprint"], leading to recursion and hence to
			# yaml parser errors

			configured_checks = self._settings.get(["checks"], incl_defaults=False)
			if configured_checks is None:
				return

			if "octoprint" in configured_checks and "octoprint" in configured_checks["octoprint"]:
				# that's a circular reference, back to defaults
				dummy_defaults = dict(plugins=dict())
				dummy_defaults["plugins"][self._identifier] = dict(checks=dict())
				dummy_defaults["plugins"][self._identifier]["checks"]["octoprint"] = None
				self._settings.set(["checks", "octoprint"], None, defaults=dummy_defaults)

		if current is None or current < 7:
			# remove check_providers again
			self._settings.set(["check_providers"], None, defaults=dict(check_providers=dict()))

	def _clean_settings_check(self, key, data, defaults, delete=None, save=True):
		if not data:
			# nothing to do
			return data

		if delete is None:
			delete = []

		for k, v in data.items():
			if k in defaults and defaults[k] == data[k]:
				del data[k]

		for k in delete:
			if k in data:
				del data[k]

		dummy_defaults = dict(plugins=dict())
		dummy_defaults["plugins"][self._identifier] = dict(checks=dict())
		dummy_defaults["plugins"][self._identifier]["checks"][key] = defaults
		if len(data):
			self._settings.set(["checks", key], data, defaults=dummy_defaults)
		else:
			self._settings.set(["checks", key], None, defaults=dummy_defaults)

		if save:
			self._settings.save()

		return data

	#~~ BluePrint API

	@octoprint.plugin.BlueprintPlugin.route("/check", methods=["GET"])
	@no_firstrun_access
	@Permissions.PLUGIN_SOFTWAREUPDATE_CHECK.require(403)
	def check_for_update(self):
		if "check" in flask.request.values:
			check_targets = list(map(lambda x: x.strip(), flask.request.values["check"].split(",")))
		else:
			check_targets = None

		force = flask.request.values.get("force", "false") in octoprint.settings.valid_boolean_trues

		def view():
			self._environment_ready.wait(timeout=30.0)

			try:
				information, update_available, update_possible = self.get_current_versions(check_targets=check_targets, force=force)
				return flask.jsonify(dict(status="inProgress" if self._update_in_progress else "updatePossible" if update_available and update_possible and self._environment_supported else "updateAvailable" if update_available else "current",
				                          information=information,
				                          timestamp=self._version_cache_timestamp,
				                          environment=dict(supported=self._environment_supported,
				                                           versions=[dict(name=gettext("Python"), current=self._environment_versions.get("python", "unknown"), minimum=MINIMUM_PYTHON),
				                                                     dict(name=gettext("pip"), current=self._environment_versions.get("pip", "unknown"), minimum=MINIMUM_PIP),
				                                                     dict(name=gettext("setuptools"), current=self._environment_versions.get("setuptools", "unknown"), minimum=MINIMUM_SETUPTOOLS)])))
			except exceptions.ConfigurationInvalid as e:
				return flask.make_response("Update not properly configured, can't proceed: %s" % e.message, 500)

		def etag():
			checks = self._get_configured_checks()

			targets = check_targets
			if targets is None:
				targets = checks.keys()

			hash = hashlib.sha1()
			def hash_update(value):
				value = value.encode('utf-8')
				hash.update(value)

			targets = sorted(targets)
			for target in targets:
				current_hash = self._get_check_hash(checks.get(target, dict()))
				if target in self._version_cache and not force:
					data = self._version_cache[target]
					hash_update(current_hash)
					hash_update(str(data["timestamp"] + self._version_cache_ttl >= time.time() > data["timestamp"]))
					hash_update(repr(data["information"]))
					hash_update(str(data["available"]))
					hash_update(str(data["possible"]))
					hash_update(str(data.get("online", None)))

			hash_update(",".join(targets))

			hash_update(str(self._environment_supported))
			hash_update(str(self._version_cache_timestamp))
			hash_update(str(self._connectivity_checker.online))
			hash_update(str(self._update_in_progress))
			hash_update(self.DATA_FORMAT_VERSION)
			return hash.hexdigest()

		def condition():
			return check_etag(etag())

		return with_revalidation_checking(etag_factory=lambda *args, **kwargs: etag(),
		                                  condition=lambda *args, **kwargs: condition(),
		                                  unless=lambda: force)(view)()


	@octoprint.plugin.BlueprintPlugin.route("/update", methods=["POST"])
	@no_firstrun_access
	@Permissions.PLUGIN_SOFTWAREUPDATE_UPDATE.require(403)
	def perform_update(self):
		throttled = self._get_throttled()
		if throttled and isinstance(throttled, dict) and throttled.get("current_issue", False) and not self._settings.get_boolean(["ignore_throttled"]):
			# currently throttled, we refuse to run
			return flask.make_response("System is currently throttled, refusing to update "
			                           "anything due to possible stability issues", 409)

		if self._printer.is_printing() or self._printer.is_paused():
			# do not update while a print job is running
			return flask.make_response("Printer is currently printing or paused", 409)

		if not self._environment_supported:
			return flask.make_response("Direct updates are not supported in this Python environment", 409)

		if not "application/json" in flask.request.headers["Content-Type"]:
			return flask.make_response("Expected content-type JSON", 400)

		json_data = flask.request.get_json(silent=True)
		if json_data is None:
			return flask.make_response("Invalid JSON", 400)

		if "targets" in json_data:
			targets = list(map(lambda x: x.strip(), json_data["targets"]))
		else:
			targets = None

		if "force" in json_data and json_data["force"] in octoprint.settings.valid_boolean_trues:
			force = True
		else:
			force = False

		to_be_checked, checks = self.perform_updates(targets=targets, force=force)
		return flask.jsonify(dict(order=to_be_checked, checks=checks))

	def is_blueprint_protected(self):
		return False

	#~~ Asset API

	def get_assets(self):
		return dict(
			css=["css/softwareupdate.css"],
			js=["js/softwareupdate.js"],
			clientjs=["clientjs/softwareupdate.js"],
			less=["less/softwareupdate.less"]
		)

	##~~ TemplatePlugin API

	def get_template_configs(self):
		from flask_babel import gettext
		return [
			dict(type="settings", name=gettext("Software Update"))
		]

	##~~ WizardPlugin API

	def is_wizard_required(self):
		checks = self._get_configured_checks()
		check = checks.get("octoprint", None)
		checkout_folder = self._get_octoprint_checkout_folder(checks=checks)
		return check and "method" in check and check["method"] == "update_script" and not checkout_folder

	##~~ EventHandlerPlugin API

	def on_event(self, event, payload):
		from octoprint.events import Events
		if event != Events.CONNECTIVITY_CHANGED or not payload or not payload.get("new", False):
			return

		thread = threading.Thread(target=self.get_current_versions)
		thread.daemon = True
		thread.start()

	#~~ Updater

	def get_current_versions(self, check_targets=None, force=False):
		"""
		Retrieves the current version information for all defined check_targets. Will retrieve information for all
		available targets by default.

		:param check_targets: an iterable defining the targets to check, if not supplied defaults to all targets
		"""

		checks = self._get_configured_checks()
		if check_targets is None:
			check_targets = list(checks.keys())

		update_available = False
		update_possible = False
		information = dict()

		# we don't want to do the same work twice, so let's use a lock
		if self._get_versions_mutex.acquire(False):
			self._get_versions_data_ready.clear()
			try:
				futures_to_result = dict()
				online = self._connectivity_checker.check_immediately()
				self._logger.debug("Looks like we are {}".format("online" if online else "offline"))

				with futures.ThreadPoolExecutor(max_workers=5) as executor:
					for target, check in checks.items():
						if not target in check_targets:
							continue

						if not check:
							continue

						try:
							populated_check = self._populated_check(target, check)
							future = executor.submit(self._get_current_version, target, populated_check, force=force)
							futures_to_result[future] = (target, populated_check)
						except exceptions.UnknownCheckType:
							self._logger.warning("Unknown update check type for target {}: {}".format(target,
							                                                                       check.get("type",
							                                                                                 "<n/a>")))
							continue
						except Exception:
							self._logger.exception("Could not check {} for updates".format(target))
							continue

					for future in futures.as_completed(futures_to_result):

						target, populated_check = futures_to_result[future]
						if future.exception() is not None:
							self._logger.error("Could not check {} for updates, error: {!r}".format(target,
							                                                                        future.exception()))
							continue

						target_information, target_update_available, target_update_possible, target_online, target_error = future.result()
						target_update_possible = target_update_possible and self._environment_supported

						target_information = dict_merge(dict(local=dict(name="?", value="?"),
						                                     remote=dict(name="?", value="?",
						                                                 release_notes=None),
						                                     needs_online=True), target_information)

						update_available = update_available or target_update_available
						update_possible = update_possible or (target_update_possible and target_update_available)

						local_name = target_information["local"]["name"]
						local_value = target_information["local"]["value"]

						release_notes = None
						if target_information and target_information["remote"] and target_information["remote"][
							"value"]:
							if "release_notes" in populated_check and populated_check["release_notes"]:
								release_notes = populated_check["release_notes"]
							elif "release_notes" in target_information["remote"]:
								release_notes = target_information["remote"]["release_notes"]

							if release_notes:
								release_notes = release_notes.format(octoprint_version=VERSION,
								                                     target_name=target_information["remote"]["name"],
								                                     target_version=target_information["remote"]["value"])

						information[target] = dict(updateAvailable=target_update_available,
						                           updatePossible=target_update_possible,
						                           information=target_information,
						                           displayName=populated_check["displayName"],
						                           displayVersion=populated_check["displayVersion"].format(octoprint_version=VERSION,
						                                                                                   local_name=local_name,
						                                                                                   local_value=local_value),
						                           releaseNotes=release_notes,
						                           online=target_online,
						                           error=target_error)

						if target == "octoprint" and "released_version" in populated_check:
							information[target]["released_version"] = populated_check["released_version"]

				if self._version_cache_dirty:
					self._save_version_cache()

				self._get_versions_data = information, update_available, update_possible
				self._get_versions_data_ready.set()
			finally:
				self._get_versions_mutex.release()

		else: # something's already in progress, let's wait for it to complete and use its result
			self._get_versions_data_ready.wait()
			information, update_available, update_possible = self._get_versions_data

		return information, update_available, update_possible

	def _get_check_hash(self, check):
		def dict_to_sorted_repr(d):
			lines = []
			for key in sorted(d.keys()):
				value = d[key]
				if isinstance(value, dict):
					lines.append("{!r}: {}".format(key, dict_to_sorted_repr(value)))
				else:
					lines.append("{!r}: {!r}".format(key, value))

			return "{" + ", ".join(lines) + "}"

		hash = hashlib.md5()
		def hash_update(value):
			value = value.encode('utf-8')
			hash.update(value)
		hash_update(dict_to_sorted_repr(check))
		return hash.hexdigest()

	def _get_current_version(self, target, check, force=False, online=None):
		"""
		Determines the current version information for one target based on its check configuration.
		"""

		current_hash = self._get_check_hash(check)
		if online is None:
			online = self._connectivity_checker.online
		if target in self._version_cache and not force:
			data = self._version_cache[target]
			if data["hash"] == current_hash \
					and data["timestamp"] + self._version_cache_ttl >= time.time() > data["timestamp"] \
					and data.get("online", None) == online:
				# we also check that timestamp < now to not get confused too much by clock changes
				return data["information"], data["available"], data["possible"], data["online"], data.get("error", None)

		information = dict()
		update_available = False
		error = None

		try:
			version_checker = self._get_version_checker(target, check)
			information, is_current = version_checker.get_latest(target, check, online=online)
			if information is not None and not is_current:
				update_available = True
		except exceptions.CannotCheckOffline:
			update_possible = False
			information["needs_online"] = True
		except exceptions.UnknownCheckType:
			self._logger.warning("Unknown check type %s for %s" % (check["type"], target))
			update_possible = False
			error = "unknown_check"
		except exceptions.NetworkError:
			self._logger.warning("Could not check %s for updates due to a network error" % target)
			update_possible = False
			error = "network"
		except Exception:
			self._logger.exception("Could not check %s for updates" % target)
			update_possible = False
			error = "unknown"
		else:
			try:
				updater = self._get_updater(target, check)
				update_possible = updater.can_perform_update(target, check, online=online)
			except Exception:
				update_possible = False

		self._version_cache[target] = dict(timestamp=time.time(),
		                                   hash=current_hash,
		                                   information=information,
		                                   available=update_available,
		                                   possible=update_possible,
		                                   online=online,
		                                   error=error)
		self._version_cache_dirty = True
		return information, update_available, update_possible, online, error

	def perform_updates(self, force=False, **kwargs):
		"""
		Performs the updates for the given check_targets. Will update all possible targets by default.

		:param targets: an iterable defining the targets to update, if not supplied defaults to all targets
		"""

		if not self._environment_supported:
			self._logger.error("Direct updates are unsupported in this environment")
			return [], dict()

		targets = kwargs.get("targets", kwargs.get("check_targets", None))

		checks = self._get_configured_checks()
		populated_checks = dict()
		for target, check in checks.items():
			try:
				populated_checks[target] = self._populated_check(target, check)
			except exceptions.UnknownCheckType:
				self._logger.debug("Ignoring unknown check type for target {}".format(target))
			except Exception:
				self._logger.exception("Error while populating check prior to update for target {}".format(target))

		if targets is None:
			targets = populated_checks.keys()
		to_be_updated = sorted(set(targets) & set(populated_checks.keys()))
		if "octoprint" in to_be_updated:
			to_be_updated.remove("octoprint")
			tmp = ["octoprint"] + to_be_updated
			to_be_updated = tmp
		if "pip" in to_be_updated:
			to_be_updated.remove("pip")
			tmp = ["pip"] + to_be_updated
			to_be_updated = tmp

		updater_thread = threading.Thread(target=self._update_worker, args=(populated_checks, to_be_updated, force))
		updater_thread.daemon = False
		updater_thread.start()

		check_data = dict((key, check["displayName"] if "displayName" in check else key) for key, check in populated_checks.items() if key in to_be_updated)
		return to_be_updated, check_data

	def _update_worker(self, checks, check_targets, force):

		restart_type = None

		try:
			self._update_in_progress = True

			target_results = dict()
			error = False

			### iterate over all configured targets

			for target in check_targets:
				if not target in checks:
					continue
				check = checks[target]

				if "enabled" in check and not check["enabled"]:
					continue

				if not target in check_targets:
					continue

				target_error, target_result = self._perform_update(target, check, force)
				error = error or target_error
				if target_result is not None:
					target_results[target] = target_result

					if "restart" in check:
						target_restart_type = check["restart"]
					elif "pip" in check:
						target_restart_type = "octoprint"
					else:
						target_restart_type = None

					# if our update requires a restart we have to determine which type
					if restart_type is None or (restart_type == "octoprint" and target_restart_type == "environment"):
						restart_type = target_restart_type

		finally:
			# we might have needed to update the config, so we'll save that now
			self._settings.save()

			# also, we are now longer updating
			self._update_in_progress = False

		if error:
			# if there was an unignorable error, we just return error
			self._send_client_message("error", dict(results=target_results))

		else:
			self._save_version_cache()

			# otherwise the update process was a success, but we might still have to restart
			if restart_type is not None and restart_type in ("octoprint", "environment"):
				# one of our updates requires a restart of either type "octoprint" or "environment". Let's see if
				# we can actually perform that

				restart_command = None
				if restart_type == "octoprint":
					restart_command = self._settings.global_get(["server", "commands", "serverRestartCommand"])
				elif restart_type == "environment":
					restart_command = self._settings.global_get(["server", "commands", "systemRestartCommand"])

				if restart_command:
					self._send_client_message("restarting", dict(restart_type=restart_type, results=target_results))
					try:
						self._perform_restart(restart_command)
					except exceptions.RestartFailed:
						self._send_client_message("restart_failed", dict(restart_type=restart_type, results=target_results))
				else:
					# we don't have this restart type configured, we'll have to display a message that a manual
					# restart is needed
					self._send_client_message("restart_manually", dict(restart_type=restart_type, results=target_results))
			else:
				self._send_client_message("success", dict(results=target_results))

	def _perform_update(self, target, check, force):
		online = self._connectivity_checker.online

		information, update_available, update_possible, _, _ = self._get_current_version(target, check, online=online)

		if not update_available and not force:
			return False, None

		if not update_possible:
			self._logger.warning("Cannot perform update for %s, update type is not fully configured" % target)
			return False, None

		# determine the target version to update to
		target_version = information["remote"]["value"]
		target_error = False
		target_result = None

		def trigger_event(success, **additional_payload):
			from octoprint.events import Events

			if success:
				# noinspection PyUnresolvedReferences
				event = Events.PLUGIN_SOFTWAREUPDATE_UPDATE_SUCCEEDED
			else:
				# noinspection PyUnresolvedReferences
				event = Events.PLUGIN_SOFTWAREUPDATE_UPDATE_FAILED

			payload = copy.copy(additional_payload)
			payload.update(dict(target=target,
			                    from_version=information["local"]["value"],
			                    to_version=target_version))

			self._event_bus.fire(event, payload=payload)

		### The actual update procedure starts here...

		populated_check = self._populated_check(target, check)
		try:
			self._logger.info("Starting update of %s to %s..." % (target, target_version))
			self._send_client_message("updating", dict(target=target, version=target_version, name=populated_check["displayName"]))
			updater = self._get_updater(target, check)
			if updater is None:
				raise exceptions.UnknownUpdateType()

			update_result = updater.perform_update(target, populated_check, target_version, log_cb=self._log, online=online)
			target_result = ("success", update_result)
			self._logger.info("Update of %s to %s successful!" % (target, target_version))
			trigger_event(True)

		except exceptions.UnknownUpdateType:
			self._logger.warning("Update of %s can not be performed, unknown update type" % target)
			self._send_client_message("update_failed", dict(target=target, version=target_version, name=populated_check["displayName"], reason="Unknown update type"))
			return False, None

		except exceptions.CannotUpdateOffline:
			self._logger.warning("Update of %s can not be performed, it's not marked as 'offline' capable but we are apparently offline right now" % target)
			self._send_client_message("update_failed", dict(target=target, version=target_version, name=populated_check["displayName"], reason="No internet connection"))

		except Exception as e:
			self._logger.exception("Update of %s can not be performed, please also check plugin_softwareupdate_console.log for possible causes of this" % target)
			trigger_event(False)

			if not "ignorable" in populated_check or not populated_check["ignorable"]:
				target_error = True

			if isinstance(e, exceptions.UpdateError):
				target_result = ("failed", e.data)
				self._send_client_message("update_failed", dict(target=target, version=target_version, name=populated_check["displayName"], reason=e.data))
			else:
				target_result = ("failed", None)
				self._send_client_message("update_failed", dict(target=target, version=target_version, name=populated_check["displayName"], reason="unknown"))

		else:
			# make sure that any external changes to config.yaml are loaded into the system
			self._settings.load()

			# persist the new version if necessary for check type
			if check["type"] in self.COMMIT_TRACKING_TYPES:
				dummy_default = dict(plugins=dict())
				dummy_default["plugins"][self._identifier] = dict(checks=dict())
				dummy_default["plugins"][self._identifier]["checks"][target] = dict(current=None)
				self._settings.set(["checks", target, "current"], target_version, defaults=dummy_default)

				# we have to save here (even though that makes us save quite often) since otherwise the next
				# load will overwrite our changes we just made
				self._settings.save()

			del self._version_cache[target]
			self._version_cache_dirty = True

		return target_error, target_result

	def _perform_restart(self, restart_command):
		"""
		Performs a restart using the supplied restart_command.
		"""

		self._logger.info("Restarting...")
		try:
			util.execute(restart_command, evaluate_returncode=False, do_async=True)
		except exceptions.ScriptError as e:
			self._logger.exception("Error while restarting via command {}".format(restart_command))
			self._logger.warning("Restart stdout:\n{}".format(e.stdout))
			self._logger.warning("Restart stderr:\n{}".format(e.stderr))
			raise exceptions.RestartFailed()

	def _populated_check(self, target, check):
		from flask_babel import gettext

		if not "type" in check:
			raise exceptions.UnknownCheckType()

		result = dict(check)

		if target == "octoprint":

			from octoprint.util.version import is_released_octoprint_version, is_stable_octoprint_version

			displayName = check.get("displayName")
			if displayName is None:
				# displayName missing or set to None
				displayName = gettext("OctoPrint")
			result["displayName"] = to_unicode(displayName, errors="replace")

			displayVersion = check.get("displayVersion")
			if displayVersion is None:
				# displayVersion missing or set to None
				displayVersion = "{octoprint_version}"
			result["displayVersion"] = to_unicode(displayVersion, errors="replace")

			stable_branch = "master"
			release_branches = []
			if "stable_branch" in check:
				release_branches.append(check["stable_branch"]["branch"])
				stable_branch = check["stable_branch"]["branch"]
			if "prerelease_branches" in check:
				release_branches += [x["branch"] for x in check["prerelease_branches"]]
			result["released_version"] = is_released_octoprint_version()

			if check["type"] in self.COMMIT_TRACKING_TYPES:
				result["current"] = REVISION if REVISION else "unknown"
			else:
				result["current"] = VERSION

				if check["type"] == "github_release" and (check.get("prerelease", None) or not is_stable_octoprint_version()):
					# we are tracking github releases and are either also tracking prerelease OR are currently running
					# a non stable version => we need to change some parameters

					# we compare versions fully, not just the base so that we see a difference
					# between RCs + stable for the same version release
					result["force_base"] = False

					if check.get("prerelease", None):
						# we are tracking prereleases => we want to be on the correct prerelease channel/branch
						channel = check.get("prerelease_channel", None)
						if channel:
							# if we have a release channel, we also set our update_branch here to our release channel
							# in case it's not already set
							result["update_branch"] = check.get("update_branch", channel)

					else:
						# we are not tracking prereleases, but aren't on the stable branch either => switch back
						# to stable branch on update
						result["update_branch"] = check.get("update_branch", stable_branch)

					if check.get("update_script", None):
						# we force an exact version & python unequality check, to be able to downgrade
						result["force_exact_version"] = True
						result["release_compare"] = "python_unequal"
					elif check.get("pip", None):
						# we force python unequality check for pip installs, to be able to downgrade
						result["release_compare"] = "python_unequal"

		elif target == "pip":
			import pkg_resources

			displayName = check.get("displayName")
			if displayName is None:
				# displayName missing or set to None
				displayName = gettext("pip")
			result["displayName"] = to_unicode(displayName, errors="replace")

			displayVersion = check.get("displayVersion")
			if displayVersion is None:
				# displayVersion missing or set to None
				distribution = pkg_resources.get_distribution("pip")
				if distribution:
					displayVersion = distribution.version
			result["displayVersion"] = to_unicode(displayVersion, errors="replace")

		else:
			result["displayName"] = to_unicode(check.get("displayName"), errors="replace")
			if result["displayName"] is None:
				# displayName missing or None
				result["displayName"] = to_unicode(target, errors="replace")

			result["displayVersion"] = to_unicode(check.get("displayVersion", check.get("current")), errors="replace")
			if result["displayVersion"] is None:
				# displayVersion AND current missing or None
				result["displayVersion"] = "unknown"

			if check["type"] in self.COMMIT_TRACKING_TYPES:
				result["current"] = check.get("current", None)
			else:
				result["current"] = check.get("current", check.get("displayVersion", None))

		if "pip" in result:
			if not "pip_command" in check and self._settings.get(["pip_command"]) is not None:
				result["pip_command"] = self._settings.get(["pip_command"])

		return result

	def _log(self, lines, prefix=None, stream=None, strip=True):
		if strip:
			lines = list(map(lambda x: x.strip(), lines))

		self._send_client_message("loglines", data=dict(loglines=[dict(line=line, stream=stream) for line in lines]))
		for line in lines:
			self._console_logger.debug("{} {}".format(prefix, line))

	def _send_client_message(self, message_type, data=None):
		self._plugin_manager.send_plugin_message(self._identifier, dict(type=message_type, data=data))

	def _get_version_checker(self, target, check):
		"""
		Retrieves the version checker to use for given target and check configuration. Will raise an UnknownCheckType
		if version checker cannot be determined.
		"""

		if not "type" in check:
			raise exceptions.ConfigurationInvalid("no check type defined")

		check_type = check["type"]
		method = getattr(version_checks, check_type)
		if method is None:
			raise exceptions.UnknownCheckType()
		else:
			return method

	def _get_update_method(self, target, check, valid_methods=None):
		"""
		Determines the update method for the given target and check.

		If ``valid_methods`` is provided, determine method must be contained
		therein to be considered valid.

		Raises an ``UnknownUpdateType`` exception if method cannot be determined
		or validated.
		"""

		method = None
		if "method" in check:
			method = check["method"]
		else:
			if "update_script" in check:
				method = "update_script"
			elif "pip" in check:
				method = "pip"
			elif "python_updater" in check:
				method = "python_updater"

		if method is None or (valid_methods and not method in valid_methods):
			raise exceptions.UnknownUpdateType()

		return method

	def _get_updater(self, target, check):
		"""
		Retrieves the updater for the given target and check configuration. Will raise an UnknownUpdateType if updater
		cannot be determined.
		"""

		mapping = dict(update_script=updaters.update_script,
		               pip=updaters.pip,
		               python_updater=updaters.python_updater,
		               sleep_a_bit=updaters.sleep_a_bit)

		method = self._get_update_method(target, check, valid_methods=list(mapping.keys()))
		return mapping[method]

	def _get_octoprint_checkout_folder(self, checks=None):
		if checks is None:
			checks = self._get_configured_checks()

		if not "octoprint" in checks:
			return None

		if "checkout_folder" in checks["octoprint"]:
			return checks["octoprint"]["checkout_folder"]
		elif "update_folder" in checks["octoprint"]:
			return checks["octoprint"]["update_folder"]

		return None

	def _get_octoprint_tracked_branch(self, checks=None):
		if checks is None:
			checks = self._get_configured_checks()

		if not "octoprint" in checks:
			return None

		return checks["octoprint"].get("branch")

	def _get_octoprint_pip_target(self, checks=None):
		if checks is None:
			checks = self._get_configured_checks()

		if not "octoprint" in checks:
			return None

		return checks["octoprint"].get("pip")


def _register_custom_events(*args, **kwargs):
	return ["update_succeeded", "update_failed"]


__plugin_name__ = "Software Update"
__plugin_author__ = "Gina Häußge"
__plugin_url__ = "http://docs.octoprint.org/en/master/bundledplugins/softwareupdate.html"
__plugin_description__ = "Allows receiving update notifications and performing updates of OctoPrint and plugins"
__plugin_disabling_discouraged__ = gettext("Without this plugin OctoPrint will no longer be able to "
                                           "update itself or any of your installed plugins which might put "
                                           "your system at risk.")
__plugin_license__ = "AGPLv3"
def __plugin_load__():
	global __plugin_implementation__
	__plugin_implementation__ = SoftwareUpdatePlugin()

	global __plugin_helpers__
	__plugin_helpers__ = dict(
		version_checks=version_checks,
		updaters=updaters,
		exceptions=exceptions,
		util=util
	)

	global __plugin_hooks__
	__plugin_hooks__ = {
		"octoprint.cli.commands": cli.commands,
		"octoprint.events.register_custom_events": _register_custom_events,
		"octoprint.access.permissions": __plugin_implementation__.get_additional_permissions
	}



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"


from .exceptions import ScriptError

import logging


def execute(command, cwd=None, evaluate_returncode=True, **kwargs):
	do_async = kwargs.get("do_async", kwargs.get("async", False))

	import sarge
	p = None

	try:
		p = sarge.run(command, cwd=cwd, stdout=sarge.Capture(), stderr=sarge.Capture(), async_=do_async)
	except Exception:
		logging.getLogger(__name__).exception("Error while executing command: {}".format(command))
		returncode = p.returncode if p is not None else None
		stdout = p.stdout.text if p is not None and p.stdout is not None else ""
		stderr = p.stderr.text if p is not None and p.stderr is not None else ""
		raise ScriptError(returncode, stdout, stderr)

	if evaluate_returncode and p.returncode != 0:
		raise ScriptError(p.returncode, p.stdout.text, p.stderr.text)

	if not do_async:
		return p.returncode, p.stdout.text, p.stderr.text



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'
__copyright__ = "Copyright (C) 2018 The OctoPrint Project - Released under terms of the AGPLv3 License"

import octoprint.plugin
from octoprint.settings import settings

from octoprint.server import NO_CONTENT
from octoprint.server.util.flask import redirect_to_tornado, no_firstrun_access
from octoprint.access import ADMIN_GROUP
from octoprint.access.permissions import Permissions

from flask import request, jsonify, url_for, make_response
from flask_babel import gettext

from werkzeug.utils import secure_filename
from werkzeug.exceptions import BadRequest
import yaml

import io
import os

try:
	from os import scandir
except ImportError:
	from scandir import scandir

class LoggingPlugin(octoprint.plugin.AssetPlugin,
                    octoprint.plugin.SettingsPlugin,
                    octoprint.plugin.TemplatePlugin,
                    octoprint.plugin.BlueprintPlugin):

	# Additional permissions hook

	def get_additional_permissions(self):
		return [
			dict(key="MANAGE",
			     name="Logging management",
			     description=gettext("Allows to download and delete log files and list and set log levels."),
			     default_groups=[ADMIN_GROUP],
			     roles=["manage"])
		]

	@octoprint.plugin.BlueprintPlugin.route("/", methods=["GET"])
	@no_firstrun_access
	@Permissions.PLUGIN_LOGGING_MANAGE.require(403)
	def get_all(self):
		files = self._getLogFiles()
		free, total = self._get_usage()
		loggers = self._get_available_loggers()
		levels = self._get_logging_levels()
		return jsonify(logs=dict(files=files, free=free, total=total),
		               setup=dict(loggers=loggers, levels=levels))

	@octoprint.plugin.BlueprintPlugin.route("/logs", methods=["GET"])
	@no_firstrun_access
	@Permissions.PLUGIN_LOGGING_MANAGE.require(403)
	def get_log_files(self):
		files = self._getLogFiles()
		free, total = self._get_usage()
		return jsonify(files=files, free=free, total=total)

	@octoprint.plugin.BlueprintPlugin.route("/logs/<path:filename>", methods=["GET"])
	@no_firstrun_access
	@Permissions.PLUGIN_LOGGING_MANAGE.require(403)
	def download_log(self, filename):
		return redirect_to_tornado(request, url_for("index") + "downloads/logs/" + filename)

	@octoprint.plugin.BlueprintPlugin.route("/logs/<path:filename>", methods=["DELETE"])
	@no_firstrun_access
	@Permissions.PLUGIN_LOGGING_MANAGE.require(403)
	def delete_log(self, filename):
		secure = os.path.join(settings().getBaseFolder("logs"), secure_filename(filename))
		if not os.path.exists(secure):
			return make_response("File not found: %s" % filename, 404)

		os.remove(secure)

		return NO_CONTENT

	@octoprint.plugin.BlueprintPlugin.route("/setup", methods=["GET"])
	@no_firstrun_access
	@Permissions.PLUGIN_LOGGING_MANAGE.require(403)
	def get_logging_setup(self):
		loggers = self._get_available_loggers()
		levels = self._get_logging_levels()
		return jsonify(loggers=loggers, levels=levels)

	@octoprint.plugin.BlueprintPlugin.route("/setup/levels", methods=["GET"])
	@no_firstrun_access
	@Permissions.PLUGIN_LOGGING_MANAGE.require(403)
	def get_logging_levels_api(self):
		return jsonify(self._get_logging_levels())

	@octoprint.plugin.BlueprintPlugin.route("/setup/levels", methods=["PUT"])
	@no_firstrun_access
	@Permissions.PLUGIN_LOGGING_MANAGE.require(403)
	def set_logging_levels_api(self):
		if not "application/json" in request.headers["Content-Type"]:
			return make_response("Expected content-type JSON", 400)

		try:
			json_data = request.json
		except BadRequest:
			return make_response("Malformed JSON body in request", 400)

		if not isinstance(json_data, dict):
			return make_response("Invalid log level configuration", 400)

		# TODO validate further

		self._set_logging_levels(json_data)
		return self.get_logging_levels_api()

	def is_blueprint_protected(self):
		return False

	def _get_usage(self):
		import psutil
		usage = psutil.disk_usage(settings().getBaseFolder("logs", check_writable=False))
		return usage.free, usage.total

	def _getLogFiles(self):
		files = []
		basedir = settings().getBaseFolder("logs", check_writable=False)
		for entry in scandir(basedir):
			files.append({
				"name": entry.name,
				"date": int(entry.stat().st_mtime),
				"size": entry.stat().st_size,
				"refs": {
					"resource": url_for(".download_log", filename=entry.name, _external=True),
					"download": url_for("index", _external=True) + "downloads/logs/" + entry.name
				}
			})

		return files

	def _get_available_loggers(self):
		return list(filter(lambda x: self._is_managed_logger(x), self._logger.manager.loggerDict.keys()))

	def _get_logging_file(self):
		# TODO this might not be the logging config we are actually using here (command line parameter...)
		return os.path.join(self._settings.getBaseFolder("base"), "logging.yaml")

	def _get_logging_config(self):
		logging_file = self._get_logging_file()

		config_from_file = {}
		if os.path.exists(logging_file) and os.path.isfile(logging_file):
			import yaml
			with io.open(logging_file, 'rt', encoding='utf-8') as f:
				config_from_file = yaml.safe_load(f)
		return config_from_file

	def _get_logging_levels(self):
		config = self._get_logging_config()
		if config is None or not isinstance(config, dict):
			return dict()

		return dict((key, value.get("level"))
		            for key, value in config.get("loggers", dict()).items()
		            if isinstance(value, dict) and "level" in value)

	def _set_logging_levels(self, new_levels):
		import logging

		config = self._get_logging_config()

		# clear all configured logging levels
		if "loggers" in config:
			purge = []
			for component, data in config["loggers"].items():
				if not self._is_managed_logger(component): continue
				try:
					del data["level"]
					self._logger.manager.loggerDict[component].setLevel(logging.INFO)
				except KeyError:
					pass
				if len(data) == 0:
					purge.append(component)
			for component in purge:
				del config["loggers"][component]
		else:
			config["loggers"] = dict()

		# update all logging levels
		for logger, level in new_levels.items():
			if logger not in config["loggers"]:
				config["loggers"][logger] = dict()
			config["loggers"][logger]["level"] = level

		# delete empty entries
		config["loggers"] = {k: v for k, v in config["loggers"].items() if len(v)}

		# save
		with octoprint.util.atomic_write(self._get_logging_file(), mode='wt', max_permissions=0o666) as f:
			yaml.safe_dump(config, f, default_flow_style=False, indent=2, allow_unicode=True)

		# set runtime logging levels now
		for logger, level in new_levels.items():
			level = logging.getLevelName(level)

			self._logger.info("Setting logger {} level to {}".format(logger, level))
			self._logger.manager.loggerDict[logger].setLevel(level)

	def _is_managed_logger(self, logger):
		return logger and (logger.startswith("octoprint") or logger.startswith("tornado"))

	def get_template_configs(self):
		return [
			dict(type="settings", custom_bindings=True)
		]

	def get_assets(self):
		return dict(js=["js/logging.js"],
		            clientjs=["clientjs/logging.js"],
		            less=["less/logging.less"],
		            css=["css/logging.css"])

__plugin_name__ = "Logging"
__plugin_author__ = "Shawn Bruce, based on work by Gina Häußge and Marc Hannappel"
__plugin_description__ = "Provides access to OctoPrint's logs and logging configuration."
__plugin_disabling_discouraged__ = gettext("Without this plugin you will no longer be able to retrieve "
                                           "OctoPrint's logs or modify the current logging levels through "
                                           "the web interface.")
def __plugin_load__():
	global __plugin_implementation__
	__plugin_implementation__ = LoggingPlugin()

	global __plugin_hooks__
	__plugin_hooks__ = {
		"octoprint.access.permissions": __plugin_implementation__.get_additional_permissions
	}



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import flask
import threading
import os
import yaml
import io
import time
import uuid
from collections import defaultdict
from flask_babel import gettext

import octoprint.plugin
from octoprint.settings import valid_boolean_trues
from octoprint.server.util.flask import restricted_access, no_firstrun_access
from octoprint.server import NO_CONTENT, current_user, admin_permission
from octoprint.util import atomic_write, monotonic_time, ResettableTimer, generate_api_key

from octoprint.access import ADMIN_GROUP
from octoprint.access.permissions import Permissions

CUTOFF_TIME = 10 * 60 # 10min
POLL_TIMEOUT = 5 # 5 seconds

class AppAlreadyExists(Exception):
	pass


class PendingDecision(object):
	def __init__(self, app_id, app_token, user_id, user_token, timeout_callback=None):
		self.app_id = app_id
		self.app_token = app_token
		self.user_id = user_id
		self.user_token = user_token
		self.created = monotonic_time()

		if callable(timeout_callback):
			self.poll_timeout = ResettableTimer(POLL_TIMEOUT, timeout_callback, [user_token])
			self.poll_timeout.start()

	def external(self):
		return dict(app_id=self.app_id,
		            user_id=self.user_id,
		            user_token=self.user_token)

	def __repr__(self):
		return u"PendingDecision({!r}, {!r}, {!r}, {!r}, timeout_callback=...)".format(self.app_id,
		                                                                               self.app_token,
		                                                                               self.user_id,
		                                                                               self.user_token)


class ReadyDecision(object):
	def __init__(self, app_id, app_token, user_id):
		self.app_id = app_id
		self.app_token = app_token
		self.user_id = user_id

	@classmethod
	def for_pending(cls, pending, user_id):
		return cls(pending.app_id, pending.app_token, user_id)

	def __repr__(self):
		return u"ReadyDecision({!r}, {!r}, {!r})".format(self.app_id,
		                                                 self.app_token,
		                                                 self.user_id)


class ActiveKey(object):
	def __init__(self, app_id, api_key, user_id):
		self.app_id = app_id
		self.api_key = api_key
		self.user_id = user_id

	def external(self):
		return dict(app_id=self.app_id,
		            api_key=self.api_key,
		            user_id=self.user_id)

	def internal(self):
		return dict(app_id=self.app_id,
		            api_key=self.api_key)

	@classmethod
	def for_internal(cls, internal, user_id):
		return cls(internal["app_id"], internal["api_key"], user_id)

	def __repr__(self):
		return u"ActiveKey({!r}, {!r}, {!r})".format(self.app_id,
		                                             self.api_key,
		                                             self.user_id)


class AppKeysPlugin(octoprint.plugin.AssetPlugin,
                    octoprint.plugin.BlueprintPlugin,
                    octoprint.plugin.SimpleApiPlugin,
                    octoprint.plugin.TemplatePlugin):

	def __init__(self):
		self._pending_decisions = []
		self._pending_lock = threading.RLock()

		self._ready_decisions = []
		self._ready_lock = threading.RLock()

		self._keys = defaultdict(list)
		self._keys_lock = threading.RLock()

		self._key_path = None

	def initialize(self):
		self._key_path = os.path.join(self.get_plugin_data_folder(), "keys.yaml")
		self._load_keys()

	# Additional permissions hook

	def get_additional_permissions(self):
		return [
			dict(key="ADMIN",
			     name="Admin access",
			     description=gettext("Allows administrating all application keys"),
			     roles=["admin"],
			     dangerous=True,
			     default_groups=[ADMIN_GROUP])
		]

	##~~ TemplatePlugin

	def get_template_configs(self):
		return [dict(type="usersettings", name=gettext("Application Keys")),
		        dict(type="settings", name=gettext("Application Keys"))]

	##~~ AssetPlugin

	def get_assets(self):
		return dict(js=["js/appkeys.js"],
		            clientjs=["clientjs/appkeys.js"],
		            less=["less/appkeys.less"],
		            css=["css/appkeys.css"])

	##~~ BlueprintPlugin mixin

	@octoprint.plugin.BlueprintPlugin.route("/probe", methods=["GET"])
	@no_firstrun_access
	def handle_probe(self):
		return NO_CONTENT

	@octoprint.plugin.BlueprintPlugin.route("/request", methods=["POST"])
	@no_firstrun_access
	def handle_request(self):
		data = flask.request.json
		if data is None:
			return flask.make_response("Missing key request", 400)

		if not "app" in data:
			return flask.make_response("No app name provided", 400)

		app_name = data["app"]
		user_id = None
		if "user" in data and data["user"]:
			user_id = data["user"]

		app_token, user_token = self._add_pending_decision(app_name, user_id=user_id)

		self._plugin_manager.send_plugin_message(self._identifier, dict(type="request_access",
		                                                                app_name=app_name,
		                                                                user_token=user_token,
		                                                                user_id=user_id))
		response = flask.jsonify(app_token=app_token)
		response.status_code = 201
		response.headers["Location"] = flask.url_for(".handle_decision_poll", app_token=app_token, _external=True)
		return response

	@octoprint.plugin.BlueprintPlugin.route("/request/<app_token>")
	@no_firstrun_access
	def handle_decision_poll(self, app_token):
		result = self._get_pending_by_app_token(app_token)
		if result:
			for pending_decision in result:
				pending_decision.poll_timeout.reset()

			response = flask.jsonify(message="Awaiting decision")
			response.status_code = 202
			return response

		result = self._get_decision(app_token)
		if result:
			return flask.jsonify(api_key=result)

		return flask.abort(404)

	@octoprint.plugin.BlueprintPlugin.route("/decision/<user_token>", methods=["POST"])
	@restricted_access
	def handle_decision(self, user_token):
		data = flask.request.json
		if not "decision" in data:
			return flask.make_response("No decision provided", 400)
		decision = data["decision"] in valid_boolean_trues
		user_id = current_user.get_name()

		result = self._set_decision(user_token, decision, user_id)
		if not result:
			return flask.abort(404)

		# Close access_request dialog for this request on all open OctoPrint connections
		self._plugin_manager.send_plugin_message(self._identifier, dict(
			type="end_request",
			user_token=user_token
		))

		return NO_CONTENT

	def is_blueprint_protected(self):
		return False # No API key required to request API access

	##~~ SimpleApiPlugin mixin

	def get_api_commands(self):
		return dict(generate=["app"],
		            revoke=["key"])

	def on_api_get(self, request):
		user_id = current_user.get_name()
		if not user_id:
			return flask.abort(403)

		if request.values.get("all") in valid_boolean_trues and Permissions.PLUGIN_APPKEYS_ADMIN.can():
			keys = self._all_api_keys()
		else:
			keys = self._api_keys_for_user(user_id)

		return flask.jsonify(keys=list(map(lambda x: x.external(), keys)),
		                     pending=dict((x.user_token, x.external()) for x in self._get_pending_by_user_id(user_id)))

	def on_api_command(self, command, data):
		user_id = current_user.get_name()
		if not user_id:
			return flask.abort(403)

		if command == "revoke":
			api_key = data.get("key")
			if not api_key:
				return flask.abort(400)

			if not admin_permission.can():
				user_for_key = self._user_for_api_key(api_key)
				if user_for_key is None or user_for_key.user_id != user_id:
					return flask.abort(403)

			self._delete_api_key(api_key)

		elif command == "generate":
			# manual generateKey
			app_name = data.get("app")
			if not app_name:
				return flask.abort(400)

			self._add_api_key(user_id, app_name.strip())

		return NO_CONTENT

	##~~ key validator hook

	def validate_api_key(self, api_key, *args, **kwargs):
		return self._user_for_api_key(api_key)

	##~~ Helpers

	def _add_pending_decision(self, app_name, user_id=None):
		app_token = self._generate_key()
		user_token = self._generate_key()

		with self._pending_lock:
			self._remove_stale_pending()
			self._pending_decisions.append(PendingDecision(app_name, app_token, user_id, user_token,
			                                               timeout_callback=self._expire_pending))

		return app_token, user_token

	def _get_pending_by_app_token(self, app_token):
		result = []
		with self._pending_lock:
			self._remove_stale_pending()
			for data in self._pending_decisions:
				if data.app_token == app_token:
					result.append(data)
		return result

	def _get_pending_by_user_id(self, user_id):
		result = []
		with self._pending_lock:
			self._remove_stale_pending()
			for data in self._pending_decisions:
				if data.user_id == user_id or data.user_id is None:
					result.append(data)
		return result

	def _expire_pending(self, user_token):
		with self._pending_lock:
			len_before = len(self._pending_decisions)
			self._pending_decisions = list(filter(lambda x: x.user_token != user_token,
				                                  self._pending_decisions))
			len_after = len(self._pending_decisions)

			if len_after < len_before:
				self._plugin_manager.send_plugin_message(self._identifier, dict(
					type="end_request",
					user_token=user_token
				))

	def _remove_stale_pending(self):
		with self._pending_lock:
			cutoff = monotonic_time() - CUTOFF_TIME
			len_before = len(self._pending_decisions)
			self._pending_decisions = list(filter(lambda x: x.created >= cutoff,
			                                      self._pending_decisions))
			len_after = len(self._pending_decisions)
			if len_after < len_before:
				self._logger.info("Deleted {} stale pending authorization requests".format(len_before - len_after))

	def _set_decision(self, user_token, decision, user_id):
		with self._pending_lock:
			self._remove_stale_pending()
			for data in self._pending_decisions:
				if data.user_token == user_token and (data.user_id == user_id or data.user_id is None):
					pending = data
					break
			else:
				return False # not found

		if decision:
			with self._ready_lock:
				self._ready_decisions.append(ReadyDecision.for_pending(pending, user_id))

		with self._pending_lock:
			self._pending_decisions = list(filter(lambda x: x.user_token != user_token,
				                                  self._pending_decisions))

		return True

	def _get_decision(self, app_token):
		self._remove_stale_pending()

		with self._ready_lock:
			for data in self._ready_decisions:
				if data.app_token == app_token:
					decision = data
					break
			else:
				return False # not found

		api_key = self._add_api_key(decision.user_id, decision.app_id)

		with self._ready_lock:
			self._ready_decisions = list(filter(lambda x: x.app_token != app_token,
				                                self._ready_decisions))

		return api_key

	def _add_api_key(self, user_id, app_name):
		with self._keys_lock:
			for key in self._keys[user_id]:
				if key.app_id.lower() == app_name.lower():
					return key.api_key

			key = ActiveKey(app_name, self._generate_key(), user_id)
			self._keys[user_id].append(key)
			self._save_keys()
			return key.api_key

	def _delete_api_key(self, api_key):
		with self._keys_lock:
			for user_id, data in self._keys.items():
				self._keys[user_id] = list(filter(lambda x: x.api_key != api_key, data))
			self._save_keys()

	def _user_for_api_key(self, api_key):
		with self._keys_lock:
			for user_id, data in self._keys.items():
				if any(filter(lambda x: x.api_key == api_key, data)):
					if self._user_manager.enabled:
						return self._user_manager.find_user(userid=user_id)
					elif user_id == "_admin" or user_id == "dummy":
						# dummy = backwards compatible
						return self._user_manager.anonymous_user_factory()
		return None

	def _api_keys_for_user(self, user_id):
		with self._keys_lock:
			return self._keys[user_id]

	def _all_api_keys(self):
		with self._keys_lock:
			result = []
			for user_id, keys in self._keys.items():
				result += keys
		return result

	def _generate_key(self):
		return generate_api_key()

	def _load_keys(self):
		with self._keys_lock:
			if not os.path.exists(self._key_path):
				return

			try:
				with io.open(self._key_path, 'rt', encoding="utf-8", errors="strict") as f:
					persisted = yaml.safe_load(f)
			except Exception:
				self._logger.exception("Could not load application keys from {}".format(self._key_path))
				return

			if not isinstance(persisted, dict):
				return

			keys = defaultdict(list)
			for user_id, persisted_keys in persisted.items():
				keys[user_id] = [ActiveKey.for_internal(x, user_id) for x in persisted_keys]
			self._keys = keys

	def _save_keys(self):
		with self._keys_lock:
			to_persist = dict()
			for user_id, keys in self._keys.items():
				to_persist[user_id] = [x.internal() for x in keys]

			try:
				with atomic_write(self._key_path, mode='wt') as f:
					yaml.safe_dump(to_persist, f, allow_unicode=True)
			except Exception:
				self._logger.exception("Could not write application keys to {}".format(self._key_path))

__plugin_name__ = "Application Keys Plugin"
__plugin_description__ = "Implements a workflow for third party clients to obtain API keys"
__plugin_author__ = "Gina Häußge, Aldo Hoeben"
__plugin_disabling_discouraged__ = gettext("Without this plugin third party clients will no longer be able to "
                                           "obtain an API key without you manually copy-pasting it.")
__plugin_implementation__ = AppKeysPlugin()
__plugin_hooks__ = {
	"octoprint.accesscontrol.keyvalidator": __plugin_implementation__.validate_api_key,
	"octoprint.access.permissions": __plugin_implementation__.get_additional_permissions
}
